#ifndef _INCLUDE_CODE_CONTEXT_H_
#define _INCLUDE_CODE_CONTEXT_H_


#include <cassert>
#include "LocalVars.h"
#include "osstream.h"

class CfSource;
class File;
class StringLitTable;
class TreeNode;

extern ostream &endCline( ostream & );


class CodeContext : public ostringstream, public LocalVars {

public:
  explicit CodeContext( CodeContext & );
  explicit CodeContext( CodeContext &, CodeContext & );
  explicit CodeContext( CodeContext &, ostream & );
  explicit CodeContext( CfSource & );
  ~CodeContext();
  
  void pushNoLineDirectives() { curr_debug++; }
  void popNoLineDirectives() { curr_debug--; }
  void setDebugLine(TreeNode *t);
  int isValidDebug() const {
    return ((curr_debug == 0) &&
	    (curr_debug_line != -1) &&
	    (curr_debug_file != NULL));
  }
  int getCurrDebugLine() const {
    assert(isValidDebug());
    return curr_debug_line;
  }
  File *getCurrDebugFile() const {
    assert(isValidDebug());
    return curr_debug_file;
  }
  int getStartDebugLine() const {
    assert(isValidDebug());
    return start_debug_line;
  }
  File *getStartDebugFile() const {
    assert(isValidDebug());
    return start_debug_file;
  }

  static CodeContext *recover( ostream & );

  void declare( const string &name, const CtType &type) {
    if (&whereDeclare == this)
      LocalVars::declare(name, type);
    else
      whereDeclare.declare(name, type);
  }

  StringLitTable &strings;

private:
  ostream &sink;
  CodeContext *parent;
  CodeContext &whereDeclare;
  TreeNode *lastNodeSeen;
  TreeNode *firstNodeSeen;

  const int level;
  int curr_debug, curr_debug_line, start_debug_line;
  File *curr_debug_file, *start_debug_file;

  void codeGen() const;
  void codeGen( ostream & ) const;

  static const int recoverSlot;

  // forbidden
  bool operator ==( const CodeContext & ) const;
};


#endif // !_INCLUDE_CODE_CONTEXT_H_
