#include "AST.h"
#include "CodeContext.h"
#include "CtType.h"
#include "UniqueId.h"
#include "decls.h"
#include "code.h"
#include "code-util.h"


const string CodeLiteralExprNode::emitExpression( CodeContext & )
{
  return codeString();
}

const string PrimitiveLitNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return "((" + type()->cType() + ") " + literal().asString() + ")";
}


const string NullPntrNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return "(void *) 0";
}


const string ThisNode::emitExpression2( CodeContext & )
{
    // side-effect free
    return MANGLE_THIS_VAR(+);
}

const string ExprNode::simpleVar( CodeContext &context )
{
  const string var = declareTemporary( context );
  const string val = emitExpression( context );
  context << var << " = " << val << ';' << endCline;
  return var;
}

const string ObjectNode::simpleVar( CodeContext &context )
{
    return emitExpression( context );
}

const string ThisNode::simpleVar( CodeContext &context )
{
    return emitExpression( context );
}

bool ObjectNode::isSimpleLvalue()
{
    return true;
}

const string ObjectNode::emitExpression2( CodeContext &context )
{
    TreeNode *src = name()->decl()->source();
    if (isForEachPairNode(src)) {
        ForEachStmtNode *f = (ForEachStmtNode *) src->parent()->parent();
        f->needPoint() = true;
    }

    return src->emitUse( context, *this );
}

bool ObjectNode::isLocalLvalue()
{
    return true;
}

const string ObjectNode::getLvalue( CodeContext &context )
{
    // side-effect free
    return string("&") + name()->decl()->source()->emitUse( context, *this );
}

bool ThisNode::isSimpleLvalue()
{
    return true;
}

const string VarDeclNode::emitUse( LocalVars &,
				   const ObjectNode &subject ) const
{
  // side-effect free
  if (!mangledIdent) {
    // DOB: make sure we've already codeGen'd the VarDeclNode and assigned it a unique identifier

    // if not, that means someone is trying to use a variable before it has been declared
    // or more likely, the decl() pointer on the NameNode which generated this reference is incorrect
    // this can be caused by using a naive .deepClone() on an AST subtree that contains VarDeclNodes
    // and NameNodes that refer to them - unless you perform a fixup, the NameNode->decls() will 
    // still point into the original subtree, instead of to their newly cloned VarDeclNode counterparts
    // deepCloneWithCrossTreeFixup() in clone.cc does a "smart" clone that takes care of this fixup for you

    cerr << "*** Variable referencing error in VarDeclNode::emitUse()" << endl;
    cerr << "tried to generate a reference to '" << *simpName()->simpName()->ident() << "' before the declaration was generated." << endl;
    cerr << "referenced node: " << endl;
    print(cerr); cerr << endl;

    cerr << "offending reference: " << endl;
    subject.print(cerr); cerr << endl;

#if 0
    const TreeNode *t = &subject;
    int cutoff = -1; // max levels to climb, set to -1 for unlimited
    while (t && t->parent() && !isMethodDeclNode(t->parent()) && !isConstructorDeclNode(t->parent()) && cutoff) {
      t = t->parent();
      cutoff--;
      }
    cerr << "context of offending reference: " << endl;
    t->pseudoprint(cerr,1);
    cerr << "AST rep: " << endl;
    t->print(cerr,1);
#endif

    assert(0); // finally, abort
    }

  return *mangledIdent;
}


const string MethodDeclNode::emitUse( LocalVars &,
				      const ObjectNode &subject ) const
{
    // side-effect free
    return MANGLE_STACK_VAR(+, *subject.name()->ident());
}


const string ConstructorDeclNode::emitUse( LocalVars &,
					   const ObjectNode &subject ) const
{
    // side-effect free
    return MANGLE_STACK_VAR(+, *subject.name()->ident());
}
