#include <map>
#include <string>
#include "utils.h"
#include "AST.h"
#include "ClassDecl.h"
#include "CodeContext.h"
#include "CtExternal.h"
#include "CtGlobal.h"
#include "CtLocal.h"
#include "PrimitiveDecl.h"
#include "code.h"
#include "code-util.h"
#include "compiler.h"
#include "decls.h"
#include "delimit.h"
#include "lgMacro.h"
#include "UniqueId.h"

typedef map< TreeNode *, string *, less< TreeNode * > > Map;


// ThisConstructorCallNode and SuperConstructorCallNode now get removed during lowering

void StaticInitNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    // semantics??
    block()->codeGen( context );
}


void EmptyStmtNode::emitStatement( CodeContext & context)
{
}

void BlockNode::emitStatement( CodeContext &context )
{
  if (stmts()->arity() == 0)
    context << " ;" << endCline;
  else if (stmts()->arity() == 1)
    stmts()->codeGen( context );
  else {
    CodeContext subcontext( context );
    stmts()->codeGen( subcontext );
    subcontext.setDebugLine(this);
  } 
}

UniqueId gotoLabelGenerator("L");

static string *gotolabel(TreeNode *t) 
{
  static Map m;
  string *& result = m[t];
  if (result)
    return result;
  else
    return (result = new string(gotoLabelGenerator.next()));
}

void LabeledStmtNode::emitStatement( CodeContext &context )
{
  context << *gotolabel(this) << ": ;\n";
  stmt()->codeGen( context );
}

void GotoNode::emitStatement( CodeContext &context )
{
  context << "goto " << *gotolabel(destination()) << ";" << endCline;
}

void IfStmtNode::emitStatement( CodeContext &context )
{
    const string predicate = condition()->emitExpression( context );
    context << "if (" << predicate << ") ";
    {
	CodeContext thenContext( context );
	thenPart()->codeGen( thenContext );
    }
    if (!elsePart()->absent()) {
	context << "else ";
	CodeContext elseContext( context );
	elsePart()->codeGen( elseContext );
    }
}

void CodeLiteralNode::emitStatement ( CodeContext &context ) {
  context << codeString() << endCline;
}

void ThrowNode::emitStatement( CodeContext &context )
{
  string exception = expr()->emitExpression( context );  
  context << "THROW(" << exception << ");" << endCline;
}


void ParameterNode::codeGen( CodeContext & context)
{
  context.setDebugLine(this);

    // dont worry (handled by methodDeclNode)
}


void NameNode::codeGen( CodeContext &context )
{
  context.setDebugLine(this);

    if (!qualifier()->absent()) {
	qualifier()->codeGen( context );
	context << "_" << *ident();
    } else
	context << *ident();
}


void PartitionStmtNode::emitStatement( CodeContext & context)
{
}


void CheckNullNode::emitStatement( CodeContext &context )
{
  const string subject = expr()->emitExpression( context );

  context << "#if BOUNDS_CHECKING" << endCline
          << "  " << lgMacro( "CHECK_NULL", *expr()->type() )
	  << '(' << subject << ");" << endCline
          << "#endif"  << endCline;

}
