#include <AST.h>
#include <SolverInferContext.h>
#include <decls.h>
#include <template.h>


void TreeNode::inferLocalChildren( const LocalContext &context ) const
{
  const unsigned numChildren = arity();
  
  for (unsigned sweep = 0; sweep < numChildren; ++sweep)
    child( sweep )->inferLocalTree( context );
}


void TreeNode::inferLocal( const LocalContext &context )
{
  infer( context );
}


////////////////////////////////////////////////////////////////////////


void ConstructorDeclNode::inferLocalChildren( const LocalContext &context ) const
{
  LocalContext subcontext( context, *decl()->methodType() );
  TreeNode::inferLocalChildren( subcontext );
}


void MethodDeclNode::inferLocalChildren( const LocalContext &context ) const
{
  LocalContext subcontext( context, *decl()->methodType() );
  TreeNode::inferLocalChildren( subcontext );
}


void TypeDeclNode::inferLocalChildren( const LocalContext &context ) const
{
  if (!decl()->isArrayInstance())
    TreeNode::inferLocalChildren( context );
}
