#include "amlapi_endpoint.h"
#include "amlapi_node.h"
#include "amlapi_uheader.h"
#include "amlapi_const.h"

void AMLAPI_LAPI_Init () {
  lapi_hndl = (lapi_handle_t *) malloc (sizeof (lapi_handle_t));
  lapi_info = (lapi_info_t *) malloc (sizeof (lapi_info_t));
  (*lapi_info).protocol = TB3_DEV;
  LAPI_Init (lapi_hndl, lapi_info);
}

/* Must not be called before AMLAPI_LAPI_Init */
int AM_Init () {
  LAPI_Qenv (*lapi_hndl, TASK_ID, &AMLAPI_MYNODE);
  LAPI_Qenv (*lapi_hndl, NUM_TASKS, &AMLAPI_NUMNODES);
  LAPI_Gfence (*lapi_hndl);

  nextEndpointId = 0;
  pthread_mutex_init (&nextEndpointId_lock, NULL);
  return AM_OK;
}

int AM_Terminate () {
  LAPI_Term (*lapi_hndl);
  return AM_OK;
}

int AM_Poll (eb_t bundle) {
  int i;
  for (i = 0; i < AMLAPI_NUM_TASK_TO_EXEC_ON_POLL; i++) {
    if (!AMLAPI_ExecuteTaskFromBundle (bundle)) { break; }
  }
  /*printf ("%i polled %i tasks from queue.\n", AMLAPI_MYNODE, i);*/
  return i;
}

/* Given an amlapi_token, extract endpoint from the list of all endpoints */
ep_t AMLAPI_ExtractEndpointFromToken (struct amlapi_token * token) {
  ep_t destEndpoint;
  /* printf ("header handler called. Source node/endpoint: %i/%i. Dest node/endpoint: %i/%i\n", AMLAPI_MYNODE, token->sourceNode, token->sourceEndpointId, token->destNode, token->destEndpointId);*/
  /* Sanity check */
  if (AMLAPI_MYNODE != token->destNode) { printf ("Holy shit!!! Wrong delivery!!\n"); return NULL; }
  if (token->destEndpointId >= nextEndpointId) { printf ("Hey, invalid endpoint number!\n"); return NULL; }
  /* Now fetch the correct function pointer and invoke it. */
  destEndpoint = endpointsInNode [token->destEndpointId];
  if (destEndpoint == NULL) { printf ("Hey, %ith endpoint at node %i has be freed.\n", token->destEndpointId, AMLAPI_MYNODE); }
}

/* Given an endpoint and handler_t index, extract the handler from the endpoint */
amlapi_handler_t AMLAPI_ExtractHandlerFromEndpoint (ep_t destEndpoint, handler_t index) {
  amlapi_handler_t handler;
  handler = destEndpoint->handler [index];
  return handler;
}

/* lookup the en_t from the translation table of an endpoint. */
en_t lookupTranslationTable (ep_t request_endpoint, int reply_endpoint) {
  en_t endpoint;
  /* Sanity check */  
  if (request_endpoint == NULL) { printf ("Request endpoint invalid.\n"); return endpoint; }
  if (!request_endpoint->translation [reply_endpoint].inuse) { printf ("Translation table entry not in use.\n"); return endpoint; }
  endpoint = request_endpoint->translation [reply_endpoint].remoteEndpoint;
  if (endpoint.node >= AMLAPI_NUMNODES) { printf ("Endpoint has wrong destination node.\n"); return endpoint; }
  return endpoint;
}

int fillToken (struct amlapi_token * tokenPtr, ep_t request_endpoint, en_t reply_endpoint) {
  if (request_endpoint == NULL) { return AM_ERR_BAD_ARG; }
  if (tokenPtr == NULL) { return AM_ERR_BAD_ARG; }
  tokenPtr->sourceNode = AMLAPI_MYNODE;
  tokenPtr->sourceEndpointId = request_endpoint->name.endpoint_id;
  tokenPtr->destNode = reply_endpoint.node;
  tokenPtr->destEndpointId = reply_endpoint.endpoint_id;
  return AM_OK;
}

void completion_handler (lapi_handle_t * hndl, void * user_info) {
  /* printf ("%i completion handler called.\n", AMLAPI_MYNODE); */
  AMLAPI_MarkTaskReady (user_info);
  return;
}

/* Application-visible API: */
int AM_Request0 (ep_t request_endpoint, int reply_endpoint, handler_t handler) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 0, NULL, 0, FALSE);
}
int AM_Request4 (ep_t request_endpoint, int reply_endpoint, handler_t handler, int arg0, int arg1, int arg2, int arg3) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 4);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 4, argv, 0, FALSE);
}
int AM_Request8 (ep_t request_endpoint, int reply_endpoint, handler_t handler, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 8, argv, 0, FALSE);
}
int AM_RequestI0 (ep_t request_endpoint, int reply_endpoint, handler_t handler, void * source_addr, int nbytes) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 0, NULL, 0, FALSE);
}
int AM_RequestI4 (ep_t request_endpoint, int reply_endpoint, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 4);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 4, argv, 0, FALSE);
}
int AM_RequestI8 (ep_t request_endpoint, int reply_endpoint, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 8, argv, 0, FALSE);
}
int AM_RequestXfer0 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 0, NULL, dest_offset, FALSE);
}
int AM_RequestXfer4 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 4, argv, dest_offset, FALSE);
}
int AM_RequestXfer8 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 8, argv, dest_offset, FALSE);
}
int AM_RequestXferAsync0 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 0, NULL, dest_offset, TRUE);
}
int AM_RequestXferAsync4 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 4, argv, dest_offset, TRUE);
}
int AM_RequestXferAsync8 (ep_t request_endpoint, int reply_endpoint, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  return AMLAPI_SendAM (request_endpoint, reply_endpoint_name, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 8, argv, dest_offset, TRUE);
}
int AM_Reply0 (void* token, handler_t handler) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 0, NULL, 0, FALSE);
}
int AM_Reply4 (void* token, handler_t handler, int arg0, int arg1, int arg2, int arg3) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 4, argv, 0, FALSE);
}
int AM_Reply8 (void* token, handler_t handler, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, NULL, 0, AMLAPI_UHEADER_PLAIN, 8, argv, 0, FALSE);
}
int AM_ReplyI0 (void* token, handler_t handler, void * source_addr, int nbytes) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 0, NULL, 0, FALSE);
}
int AM_ReplyI4 (void* token, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 4, argv, 0, FALSE);
}
int AM_ReplyI8 (void* token, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_I, 8, argv, 0, FALSE);
}
int AM_ReplyXfer0 (void* token, int dest_offset, handler_t handler, void * source_addr, int nbytes) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 0, NULL, dest_offset, FALSE);
}
int AM_ReplyXfer4 (void* token, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 4, argv, dest_offset, FALSE);
}
int AM_ReplyXfer8 (void* token, int dest_offset, handler_t handler, void * source_addr, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  struct amlapi_token * t = (struct amlapi_token *) token;
  ep_t endpoint = AMLAPI_ExtractEndpointFromToken (t);
  en_t origin_endpoint;
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  origin_endpoint.node = t->sourceNode;
  origin_endpoint.endpoint_id = t->sourceEndpointId;
  return AMLAPI_SendAM (endpoint, origin_endpoint, handler, source_addr, nbytes, AMLAPI_UHEADER_XFER, 8, argv, dest_offset, FALSE);
}
int AM_GetXfer0 (ep_t request_endpoint, int reply_endpoint, int source_offset, handler_t handler, int dest_offset, int nbytes) {
  AM_GetXferM (request_endpoint, reply_endpoint, source_offset, handler, dest_offset, nbytes, 0, NULL);
}
int AM_GetXfer4 (ep_t request_endpoint, int reply_endpoint, int source_offset, handler_t handler, int dest_offset, int nbytes, int arg0, int arg1, int arg2, int arg3) {
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;
  AM_GetXferM (request_endpoint, reply_endpoint, source_offset, handler, dest_offset, nbytes, 4, argv);
}
int AM_GetXfer8 (ep_t request_endpoint, int reply_endpoint, int source_offset, handler_t handler, int dest_offset, int nbytes, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7) {
  int* argv = (int*) malloc (sizeof (int) * 8);
  argv [0] = arg0;  argv [1] = arg1;  argv [2] = arg2;  argv [3] = arg3;  argv [4] = arg4;  argv [5] = arg5;  argv [6] = arg6;  argv [7] = arg7;
  AM_GetXferM (request_endpoint, reply_endpoint, source_offset, handler, dest_offset, nbytes, 8, argv);
}
int AM_MaxShort () { return AMLAPI_MAX_SHORT; }
int AM_MaxMedium () { return AMLAPI_MAX_MEDIUM; }
int AM_MaxLong () { return AMLAPI_MAX_LONG; }

/* get Xfer handler. Just issue an am_Xfer */
void get_xfer_completion_handler (lapi_handle_t * hndl, void * user_info) {
  getxferhdr * xferuheader = (getxferhdr *) user_info;
  ep_t source_endpoint = AMLAPI_ExtractEndpointFromToken (&(xferuheader->token));
  int nbytes = xferuheader->nbytes;
  void * source_addr = &(((char*) source_endpoint->segAddr) [xferuheader->source_offset]);
  en_t dest_endpoint_name;
  dest_endpoint_name.node = xferuheader->token.sourceNode;
  dest_endpoint_name.endpoint_id = xferuheader->token.sourceEndpointId;
  AMLAPI_SendAM (source_endpoint, dest_endpoint_name, xferuheader->handler, source_addr, xferuheader->nbytes, AMLAPI_UHEADER_XFER, xferuheader->numArgs, xferuheader->argv, xferuheader->dest_offset, FALSE);
}
void* get_xfer_header_handler (lapi_handle_t * hndl, void * uhdr, uint * uhdr_len, unsigned long * msg_len, compl_hndlr_t ** comp_h, void ** user_info) {
  getxferhdr* xferheader = (getxferhdr *) malloc (sizeof (getxferhdr));
  memcpy (xferheader, uhdr, sizeof (getxferhdr));
  *user_info = (void*) xferheader;
  *comp_h = get_xfer_completion_handler;
}

int AM_GetXferM (ep_t request_endpoint, int reply_endpoint, int source_offset, handler_t handler, int dest_offset, int nbytes, int numArgs, int * argv) {
  lapi_cntr_t * lapi_cntr = (lapi_cntr_t *) malloc (sizeof (lapi_cntr_t));
  int val = 0;
  getxferhdr * xferuheader = (getxferhdr *) malloc (sizeof (getxferhdr));
  en_t reply_endpoint_name = lookupTranslationTable (request_endpoint, reply_endpoint);
  xferuheader->handler = handler;
  xferuheader->source_offset = source_offset;
  xferuheader->dest_offset = dest_offset;
  xferuheader->handler = handler;
  xferuheader->nbytes = nbytes;
  xferuheader->numArgs = numArgs;
  xferuheader->token.sourceNode = request_endpoint->name.node;
  xferuheader->token.sourceEndpointId = request_endpoint->name.endpoint_id;
  xferuheader->token.destNode = reply_endpoint_name.node;
  xferuheader->token.destEndpointId = reply_endpoint_name.endpoint_id;
  if (numArgs >= 1) { xferuheader->argv [1] = argv [1]; }
  if (numArgs >= 2) { xferuheader->argv [2] = argv [2]; }
  if (numArgs >= 3) { xferuheader->argv [3] = argv [3]; }
  if (numArgs >= 4) { xferuheader->argv [4] = argv [4]; }
  if (numArgs >= 5) { xferuheader->argv [5] = argv [5]; }
  if (numArgs >= 6) { xferuheader->argv [6] = argv [6]; }
  if (numArgs >= 7) { xferuheader->argv [7] = argv [7]; }
  if (numArgs >= 8) { xferuheader->argv [8] = argv [8]; }
  LAPI_Setcntr (*lapi_hndl, lapi_cntr, 0);
  LAPI_Amsend (*lapi_hndl, reply_endpoint_name.node, (void*) get_xfer_header_handler, xferuheader, sizeof (getxferhdr), NULL, 0, NULL, lapi_cntr, NULL);
  while (val == 0) { LAPI_Getcntr (*lapi_hndl, lapi_cntr, &val); }
}

void* header_handler (lapi_handle_t * hndl, void * uhdr, uint * uhdr_len, unsigned long * msg_len, compl_hndlr_t ** comp_h, void ** user_info) {
  uhdr_t uheader = (uhdr_t) uhdr;
  int nbytes = uheader->nbytes;
  int type = uheader->type;
  int M = uheader->M;
  handler_t handler_index = uheader->handler;
  int destOffset = uheader->destOffset;
  struct amlapi_token * token = (struct amlapi_token *) malloc (sizeof (struct amlapi_token));
  int * argv = (int *) malloc (M * sizeof (int)); /* freed when the AM handler is executed from the task queue */
  void* dest_addr;
  ep_t endpoint;
  amlapi_handler_t handler;

  memcpy (token, &(uheader->token), sizeof (struct amlapi_token));
  memcpy (argv, &(uheader->argv), (sizeof (int) * M));
  endpoint = AMLAPI_ExtractEndpointFromToken (token);
  handler = AMLAPI_ExtractHandlerFromEndpoint (endpoint, handler_index);

  if (type == AMLAPI_UHEADER_I) { dest_addr = (void*) malloc (nbytes); } /* freed when the AM handler is executed from the task queue */
  if (type == AMLAPI_UHEADER_XFER || type == AMLAPI_UHEADER_XFERASYNC) { 
    char* segAddr_byte_view = endpoint->segAddr;
    if (endpoint->segAddr == NULL) { printf ("Endpoint  does not have associated VM segment!! \n"); }
    if (endpoint->segLength < (destOffset + nbytes)) { printf ("Endpoint's VM segment isn't that big!! (%i + %i > %i) \n", destOffset, nbytes, endpoint->segLength); }
    dest_addr = (void*) (&(segAddr_byte_view [destOffset]));
  }
  AMLAPI_ScheduleTaskOnBundle (endpoint->endpoint_bundle, handler, token, dest_addr, nbytes, M, argv, type);
  if (type != AMLAPI_UHEADER_PLAIN) { /* If the type is plain, no need to wait for completion handler */
    *comp_h = completion_handler;
    *user_info = dest_addr;
  }
}

/* handles the common code needed to send an AM over LAPI. */
int AMLAPI_SendAM (ep_t source_endpoint, en_t dest_endpoint, handler_t handler, void * source_addr, int nbytes, int type, int M, int * argv, int dest_offset, char isAsync) {
  uhdr_t uheader;

  lapi_cntr_t * lapi_cntr = (lapi_cntr_t *) malloc (sizeof (lapi_cntr_t));
  int val = 0;

  int i;
  int filledToken;

  uheader = (uhdr_t) malloc (sizeof (struct amlapi_uheader));

  filledToken = fillToken (&(uheader->token), source_endpoint, dest_endpoint);
  if (filledToken != AM_OK) { return filledToken; }
  uheader->handler = handler;
  uheader->nbytes = nbytes;
  uheader->type = type;
  uheader->M = M;  
  uheader->destOffset = dest_offset;
  for (i = 0; i < M; i++) { uheader->argv [i] = argv [i]; }

  if (!AMLAPI_checkHeaderType (type)) { printf ("Wrong header type: %i.\n", type); return AM_ERR_BAD_ARG; }

  /* really send the AM */
  LAPI_Setcntr (*lapi_hndl, lapi_cntr, 0);
  LAPI_Amsend (*lapi_hndl, dest_endpoint.node, (void*) header_handler, uheader, sizeof (struct amlapi_uheader), (void*) source_addr, nbytes, NULL, (isAsync ? lapi_cntr : NULL), (isAsync ? NULL : lapi_cntr));
  while (val == 0) { LAPI_Getcntr (*lapi_hndl, lapi_cntr, &val); }

  free (lapi_cntr);
  free (uheader);
  return AM_OK;
}
