#ifndef _AMLAPI_ENDPOINT_H_
#define _AMLAPI_ENDPOINT_H_

#include <pthread.h>

/* End point object */

typedef uint64_t tag_t;
typedef uint8_t handler_t;
typedef void (*amlapi_handler_t) ();

#define AMLAPI_MAX_NUMTRANSLATIONS 256
#define AMLAPI_MAX_NUMHANDLERS 256
#define AMLAPI_MAX_SEGLENGTH 524288
#define AMLAPI_ENDPOINT_BUNDLE_INC 10
#define AMLAPI_ENDPOINT_BUNDLE_TASK_QUEUE_INC 10

typedef struct amlapi_ep_name {
  int node;
  int endpoint_id;
} en_t;

struct amlapi_translation_entry_t {
  char inuse;
  en_t remoteEndpoint; /* end point living on the remote node */
  tag_t tag;
  uint8_t id;
};

typedef struct amlapi_handler_q_entry {
  amlapi_handler_t handler;
  void* token;
  void* user_info;
  int nbytes;
  int numArgs;
  int * argv;
  int type;
  char ready;
} qe_t;

typedef struct amlapi_eb {
  struct amlapi_ep ** endpoints; /* dynamically-grown array of endpoints in bundle (read ep_t *)*/ 
  int size;                      /* how many endpoints are there in the bundle */
  int cursize;                   /* how much space is allocated */
  uint8_t event_mask;
  int type;
  pthread_mutex_t semLock;
  int semaphore;

  qe_t * task_queue; /* array of handler entries. Dynamically grown and shrunk. */
  int task_queue_head;
  int task_queue_tail;
  int task_queue_size;
  pthread_mutex_t task_queue_lock;
} *eb_t;

typedef struct amlapi_ep {
  eb_t endpoint_bundle; /* endpoint bundle this endpoint belongs to */
  tag_t tag;
  void *segAddr;        /* Start address of EP VM segment */
  uint32_t segLength;   /* Length of EP VM segment    */  
  en_t name;

  struct amlapi_translation_entry_t translation[AMLAPI_MAX_NUMTRANSLATIONS]; /* translation table */
  amlapi_handler_t  handler[AMLAPI_MAX_NUMHANDLERS]; /* handler table */
  char handler_inuse [AMLAPI_MAX_NUMHANDLERS]; /* checking if handler table is in use */
} *ep_t;

#endif /*_AMLAPI_ENDPOINT_H_*/
