#include "amlapi_endpoint.h"
#include "amlapi_uheader.h"
#include "amlapi_node.h"

void AMLAPI_ScheduleTaskOnBundle (eb_t bundle, amlapi_handler_t handler, void* token, void* user_info, int nbytes, int numArgs, int* argv, int type) {
  while (pthread_mutex_trylock (&(bundle->task_queue_lock))) { }  if (bundle->task_queue_tail >= bundle->task_queue_size) {
    /* either grow or move to front */
    if (bundle->task_queue_head > (bundle->task_queue_size / 2)) { /* move to front here. */
      int i;
      int j = bundle->task_queue_head;
      int size = bundle->task_queue_tail - bundle->task_queue_head;
      for (i = 0; i < size; i++) {
	bundle->task_queue [i] = bundle->task_queue [j];
	j++;
      }
      bundle->task_queue_head = 0;
      bundle->task_queue_tail = size;
    } else { /* grow the array */
      int newQueueSize = bundle->task_queue_size + AMLAPI_ENDPOINT_BUNDLE_TASK_QUEUE_INC;
      qe_t * oldQueue;
      qe_t * newQueue = (qe_t *) malloc (sizeof (qe_t) * (newQueueSize));
      int i;
      int size = bundle->task_queue_tail - bundle->task_queue_head;
      int j = bundle->task_queue_head;
      for (i = 0; i < size; i++) {
	newQueue [i] = bundle->task_queue [j];
	j++;
      }
      oldQueue = bundle->task_queue;
      bundle->task_queue = newQueue;
      bundle->task_queue_head = 0;
      bundle->task_queue_tail = size;
      bundle->task_queue_size = newQueueSize;
      free (oldQueue);
    }
  }
  bundle->task_queue [bundle->task_queue_tail].handler = handler;
  bundle->task_queue [bundle->task_queue_tail].token = token;
  bundle->task_queue [bundle->task_queue_tail].user_info = user_info;
  bundle->task_queue [bundle->task_queue_tail].nbytes = nbytes;
  bundle->task_queue [bundle->task_queue_tail].numArgs = numArgs;
  bundle->task_queue [bundle->task_queue_tail].argv = argv;
  bundle->task_queue [bundle->task_queue_tail].type = type;
  bundle->task_queue [bundle->task_queue_tail].ready = (type == AMLAPI_UHEADER_PLAIN);
  bundle->task_queue_tail++;
  pthread_mutex_unlock (&(bundle->task_queue_lock));
}

/* returns FALSE if didn't execute any task, TRUE if executed a task */
char AMLAPI_ExecuteTaskFromBundle (eb_t bundle) {
  amlapi_handler_t handler;
  void* token;
  void* user_info;
  int nbytes;
  int numArgs;
  int* argv;
  int type;

  while (pthread_mutex_trylock (&(bundle->task_queue_lock))) { }
  if (bundle->task_queue_head == bundle->task_queue_tail) {
    pthread_mutex_unlock (&(bundle->task_queue_lock));
    return FALSE;
  }
  if (!(bundle->task_queue [bundle->task_queue_head].ready)) { 
    pthread_mutex_unlock (&(bundle->task_queue_lock));
    return FALSE;
  }
  handler = (bundle->task_queue [bundle->task_queue_head]).handler;
  token = (bundle->task_queue [bundle->task_queue_head]).token;
  user_info = (bundle->task_queue [bundle->task_queue_head]).user_info;
  nbytes = (bundle->task_queue [bundle->task_queue_head]).nbytes;
  numArgs = (bundle->task_queue [bundle->task_queue_head]).numArgs;
  argv = (bundle->task_queue [bundle->task_queue_head]).argv;
  type = (bundle->task_queue [bundle->task_queue_head]).type;
  bundle->task_queue_head++;
  pthread_mutex_unlock (&(bundle->task_queue_lock));

  if (numArgs == 0 && type == AMLAPI_UHEADER_PLAIN) { handler (token); }
  if (numArgs == 4 && type == AMLAPI_UHEADER_PLAIN) { handler (token, argv [0], argv [1], argv [2], argv [3]); }
  if (numArgs == 8 && type == AMLAPI_UHEADER_PLAIN) { handler (token, argv [0], argv [1], argv [2], argv [3], argv [4], argv [5], argv [6], argv [7]); }
  if (numArgs == 0 && type == AMLAPI_UHEADER_I) { handler (token, user_info, nbytes); }
  if (numArgs == 4 && type == AMLAPI_UHEADER_I) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3]); }
  if (numArgs == 8 && type == AMLAPI_UHEADER_I) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3], argv [4], argv [5], argv [6], argv [7]); }
  if (numArgs == 0 && type == AMLAPI_UHEADER_XFER) { handler (token, user_info, nbytes); }
  if (numArgs == 4 && type == AMLAPI_UHEADER_XFER) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3]); }
  if (numArgs == 8 && type == AMLAPI_UHEADER_XFER) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3], argv [4], argv [5], argv [6], argv [7]); }
  if (numArgs == 0 && type == AMLAPI_UHEADER_XFERASYNC) { handler (token, user_info, nbytes); }
  if (numArgs == 4 && type == AMLAPI_UHEADER_XFERASYNC) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3]); }
  if (numArgs == 8 && type == AMLAPI_UHEADER_XFERASYNC) { handler (token, user_info, nbytes, argv [0], argv [1], argv [2], argv [3], argv [4], argv [5], argv [6], argv [7]); }
  free (argv);
  free (token);
  if (type == AMLAPI_UHEADER_I) { free (user_info); }
  return TRUE;
}

/* Marks the task which requires this buffer to be filled to be ready. */
void AMLAPI_MarkTaskReady (void* buffer) { 
  int i, j;
  for (i = 0; i < nextEndpointId; i++) {
    ep_t ep = endpointsInNode [i];
    eb_t bundle = ep->endpoint_bundle;
    while (pthread_mutex_trylock (&(bundle->task_queue_lock))) { }
    for (j = bundle->task_queue_head; j < bundle->task_queue_tail; j++) {
      if (bundle->task_queue [j].user_info == buffer) { /* compare the pointers */
	bundle->task_queue [j].ready = TRUE;
	pthread_mutex_unlock (&(bundle->task_queue_lock));
	return;
      }
    }
    pthread_mutex_unlock (&(bundle->task_queue_lock));
  }
  printf ("Error: Got a buffer address that is not waiting to be filled. \n");
}
