/*									tab:8
 *
 * main.c - Main program for Generic Active Message layers
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		49
 * Creation Date:	Sat Nov 20 16:47:42 1993
 * Filename:		main.c
 * History:
 *	LTL	41	Mon Jul 24 18:21:57 1995
 *		Added all_set_PROCS, which prints out an error when new_PROCS != PROCS
 *	LTL	34	Mon Jul 24 16:28:20 1995
 *		Removed #define's for the default version
 *	LTL	32	Mon Feb 20 09:36:06 1995
 *		Modified to use GAM. Used #ifdef instead of having separate platform
 *    specific versions.
 *	ssl	2	Mon Apr 18 20:54:41 1994
 *		modified for SP1
 */

#include <stdio.h>
#include <signal.h>
#include <assert.h>
#define TIC_INTERNAL_DEFINE	/* turns extern's to not externs */
#include <tic.h>

/* get MAXHOSTNAMELEN */ 
#if defined(SOLARIS)
#include <netdb.h>
#else 
#include <sys/param.h>
#endif /* SOLARIS */

#if HAVE_SCHED_YIELD
#include <sched.h>
#endif

#if defined(COMM_AM2) && defined(HAVE_ENAMESERVER_H)
#include <enameserver.h>
#endif

#if HAVE_ALLOCA_H
#include <alloca.h>
#endif

#include "main.h"
#include "ti-gc.h"

int	___AM_MAX_SHORT;
int	___AM_MAX_MEDIUM;
int	___AM_MAX_LONG;

#if defined(COMM_AM2)
/* If request and reply bundles are the same, then share variable */
eb_t	__sc_bundle;
ep_t	__sc_endpoint;
en_t    __sc_endpoint_name;
char __sc_node_description[512]; /* for debugging purposes */
#endif /* AMII */
int usingAMUDPSPMD = 0; /* true for workers using the AMUDP SPMD job startup API */

static char *progName;

int politep = 1;  /* True */


/* return log2 if exact, else return -1 */
int getLog(int i)
{
	int exp = 0;

	if(i <= 0) return -1;
	for(exp=0; (i&1)==0; i>>=1) exp++;
	return i==1 ? exp : -1;
}



#if defined(COMM_AM2) && defined(SOLARIS) && !defined(COMM_AMUDP)
#include <am2/arg_block.h>

void __error_handler(int status, op_t opcode, arg_block_t *argblock)
{
	int i;

	printf("AM layer failed on node %d.\n", myProcess());

	printf ("opcode = %d\n", argblock->opcode);
	printf ("request endpoint: %#X\n", argblock->request_endpoint);
	printf ("reply endpoint: %d\n", argblock->reply_endpoint);
	printf ("dest offset: %d\n", argblock->dest_offset);
	printf ("source addr: %#X\n", argblock->source_addr);
	printf ("nbytes: %d\n", argblock->nbytes);
	printf ("handler: %d\n", argblock->handler);
	for (i = 0; i < 8; i++)
	    printf ("data[%d]: %d\n", i, argblock->data[i]);

	switch (status) {
	  case EBADARGS:
		printf("Invalid arguements to request or reply\n");
		break;
	  case EBADENTRY:
		printf("Unbound table entry\n");
		break;
	  case EBADTAG:
		printf("Sender Tag did not match Receiver Tag\n");
		break;
	  case EBADHANDLER:
		printf("Invalid index into receiver's handler table\n");
		break;
	  case EBADSEGOFF:
		printf("Invalid offset into destination memory segment\n");
		break;
	  case EBADLENGTH:
		printf("Bulk transfer length exceeds sengment length\n");
		break;
	  case EBADENDPOINT:
		printf("Destination endpoint does not exist\n");
		break;
	  case ECONGESTION:
		printf("Persistent congestion at destination\n");
		break;
	  case EUNREACHABLE:
		printf("Unreachable destination\n");
		break;
	  default:
		printf("Unknown status value: %d\n", status);
	}

	fflush(stdout);
	exit(-1);
}
#endif /* COMM_AM2 */

#define NS_Safe(func) { if (func != 0) {fprintf(stderr, "ERROR: %s failed\n", #func); exit(1);}}


#ifdef COMM_AM2
#ifdef HAVE_ENAMESERVER_H
void setupNames()			   
{
  char 	inName[2048];
  byte	*result;
  int		result_len;
  int 	i;
  int		translations;
  en_t    *__sc_global = (en_t *) malloc(sizeof(en_t)*COMM_GetBParallelDegree());

  AM_Safe(AM_MaxNumTranslations(&translations));
  if (translations < COMM_GetBParallelDegree()) {
    fprintf(stderr,
	    "ERROR: AM_MaxNumTranslations (%d) less than BOXES (%d).\n",
	    translations, COMM_GetBParallelDegree());
    exit(-1);
  }

  /* Insert our global endpoint name into the name server */
  if (PROCS == 1) {
    AM_Safe(AM_Map(__sc_endpoint, 0, __sc_endpoint_name, TIC_TAG));
  }
  else {
#ifdef GLUNIX
    NS_Safe(NameServer_Init());
#endif
    sprintf(__sc_node_description, 
	    "tic mill progname: %s global pid: %d mybox: %d", 
	    progName, COMM_GetMyGuid(), COMM_GetMyBoxNumber());
    NS_Safe (NameServer_Bind_Version_Flags
	     (__sc_node_description, (unsigned char*)&__sc_endpoint_name, sizeof (en_t),
	      ENS_VERSION_NOT_THERE,
	      ENS_FLAG_DESTROY_BINDING_ON_CONNECTION_CLOSE));
#ifdef GLUNIX
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#endif /* Glunix */
    for (i = 0; i < COMM_GetBParallelDegree(); i++) {
      if (COMM_GetMyBoxNumber() != i) {
	sprintf(inName, "tic mill progname: %s global pid: %d mybox: %d", 
		progName, COMM_GetMyGuid(), i);
	do {
	  NameServer_Lookup(inName, &result, &result_len);
	} while (result_len == 0);
	if (result == NULL) {
	  fprintf(stderr, "ERROR: result Null for lookup %s\n", inName);
	  exit(-1);
	}			
	if (result_len != sizeof(en_t)) {
	  fprintf(stderr, 
		  "ERROR: result_len(%d)!=sizeof(en_t)(%d) for lookup %s\n",
		  result_len, sizeof(en_t), inName);
	  exit(-1);
	}
	memcpy(&__sc_global[i], result, sizeof(en_t));
	AM_Safe(AM_Map(__sc_endpoint, i, __sc_global[i], TIC_TAG));
      }
    }
  }
}

#elif defined(COMM_AMLAPI)
int setupNames() {
  int i;
  int counter = 0;
	/* Array to store en_t address from all tasks */
  void **global_addr = (void**) malloc(sizeof(void*)*COMM_GetBParallelDegree());
	/* Buffer of en_t address */
  en_t *buffer = (en_t *) malloc(sizeof(en_t));
  int* buffer_test = (int *) malloc (sizeof (int));

	/* Origin and Target Counter with address */
  void **tgt_addr = (void**) malloc(sizeof(void*)*COMM_GetBParallelDegree());
  lapi_cntr_t  lapi_cntr;	/* Origin counter */

  lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */

	/* Initialize counters to be zero at the start */
  LAPI_Setcntr(*lapi_hndl, &lapi_cntr, 0);

	/* Exchange buffer address and target counter of every task */
  LAPI_Address_init (*lapi_hndl, &__sc_endpoint_name, global_addr);

  LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
  for (i = 0; i < COMM_GetBParallelDegree(); i++) {
    if (COMM_GetMyBoxNumber() != i) {
      LAPI_Get(*lapi_hndl, i, sizeof(en_t), (void*) global_addr[i],
			(void*) buffer, NULL, &lapi_cntr); 
      while( counter == 0 ) LAPI_Getcntr(*lapi_hndl, &lapi_cntr, &counter);
      LAPI_Setcntr(*lapi_hndl, &lapi_cntr, 0);
      counter = 0;
      /* printf ("%i thinks %i's endpoint name is: %i/%i and it's own endpoint name is %i/%i \n", AMLAPI_MyProc (), i, buffer->node, buffer->endpoint_id, __sc_endpoint_name.node, __sc_endpoint_name.endpoint_id); */

      AM_Safe(AM_Map(__sc_endpoint, i, *buffer, TIC_TAG));
    } 
    LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
  }
}
#else
void setupNames()
{
    fprintf(stderr,
            "ERROR: tried to access a global endpoint name server, but Tic was compiled with no nameserver\n"); 
    fflush(stderr);
    exit(1);
}
#endif
#endif /* COMM_AM2 */

#ifdef COMM_AM2

extern void bulkSetupHandlers();

char *___AM_SEGOFFSET = NULL;

void setupAM()
{
  int 	maxSegLength;
  int 	maxHandlers;
  int		i;
	
  if (!usingAMUDPSPMD) {
    AM_Safe(AM_Init());
    AM_Safe(AM_AllocateBundle(AM_SEQ, &__sc_bundle));
    AM_Safe(AM_AllocateEndpoint(__sc_bundle, &__sc_endpoint,
			        &__sc_endpoint_name));
  }
  AM_Safe(AM_MaxSegLength(&maxSegLength));
  AM_Safe(AM_SetSeg(__sc_endpoint, (void *)DATASTART, maxSegLength));
  ___AM_SEGOFFSET = DATASTART;
	
  /*	AM_Safe(AM_SetEventMask(__sc_bundle, AM_NOEVENTS)); */

  /* Setup Handlers */
  if ((maxHandlers = AM_MaxNumHandlers()) < __LAST_HANDLER) {
    fprintf(stderr, "ERROR: Only %d handler table entries available; need %d\n",
	    maxHandlers, __LAST_HANDLER);
    exit(1);
  }
	
  for (i = 0; ; i++) {
    if (handler_list[i].handler_index == __LAST_HANDLER) {
      break;
    }
    AM_Safe(AM_SetHandler(__sc_endpoint, handler_list[i].handler_index,
			  handler_list[i].func));
  }
		
  for (i = 0; ; i++) {
    if (reply_handler_list[i].handler_index == __LAST_HANDLER) {
      break;
    }
    AM_Safe(AM_SetHandler(__sc_endpoint,
			  reply_handler_list[i].handler_index,
			  reply_handler_list[i].func));
  }

  bulkSetupHandlers();
		
  if (!usingAMUDPSPMD) 
    AM_Safe(AM_SetTag(__sc_endpoint, TIC_TAG));

  ___AM_MAX_SHORT = AM_MaxShort();
  ___AM_MAX_MEDIUM = AM_MaxMedium();
  ___AM_MAX_LONG = AM_MaxLong();
}

#endif /* COMM_AM2 */

int parseNumNodes(int *pargc, char ***pargv) {
  /* called by setup_global_layer to parse and remove the number of nodes 
   * argument for an implicit startup master 
   */
  int num_nodes;
  int i;

  if (*pargc < 2) {
    fprintf(stderr, "Tic: Missing parallel degree\n");
    fprintf(stderr, "Tic: Specify degree as first argument or use rexec/glurun/tcrun to start job\n");
    fprintf (stderr,"Tic: Usage '%s <num_nodes> {program arguments}'\n", (*pargv)[0]);
    exit(-1);
  }
  /*
   * argv[1] is number of nodes; argv[0] is program name; argv is
   * list of arguments including program name and number of nodes.
   * We need to remove argv[1] when the argument array is passed
   * to the tic_main().
   */
  num_nodes = atoi((*pargv)[1]);
  if (num_nodes < 1) {
    fprintf (stderr, "Tic: Invalid number of nodes: %s\n", (*pargv)[1]);
    fprintf (stderr, "Tic: Usage '%s <num_nodes> {program arguments}'\n", (*pargv)[0]);
    exit (1);
  }
  #ifdef GLUNIX
  { int i = Glib_GetNumAvailableNodes();
    if (i < num_nodes) {
      fprintf (stderr, "Tic: Error - Only %d nodes available (requested %d).\n",
               i, num_nodes);
      exit (1);
    }
  } 
  #endif	

  /* remove the num_nodes argument */
  for (i = 1; i < (*pargc)-1; i++) {
    (*pargv)[i] = (*pargv)[i+1];
  }
  (*pargv)[(*pargc)-1] = NULL;
  (*pargc)--;

  if (!getenv("TI_BACKEND_SILENT")) {
    printf("Tic: Spawning %s on %d nodes\n", (*pargv)[0], num_nodes);
    fflush(stdout);
  }

  return num_nodes;
}

extern julong gpid;

void setup_global_layer(int *pargc, char ***pargv) {

/* this function is responsible for performing "implicit" job startup.
 * That is, spawning a parallel job when the application was executed
 * directly without the use of a job spawner like glurun, rexec, tcrun, etc.
 * It also takes advantage of AMUDP's flexible job startup API where appropriate
 * The command-line format for all implicit startups is:
 * <programname> <paralleldegree> {program args...}
 */

#if !defined(GLUNIX) && !defined(COMM_AMUDP) && !(defined(REXEC) && defined(COMM_AMUDP))
  /* we only support implicit startup for the above job spawning systems */
  return;
#endif

/* generic initialization */
#ifdef GLUNIX
  if (!(Glib_Initialize() != 0)) {
    fprintf(stderr, "Global layer initialization failed.\n");
    fflush(stderr);
    exit(-1);
  }	
#endif /* Glunix */

#ifdef COMM_AMUDP
  AMUDP_VerboseErrors = 1; /* get detailed info about any failures */
  if (getenv("TI_BACKEND_SILENT")) {
    AMUDP_SilentMode = 1; /* suppress diagnostic output */
  }
  if (AMUDP_SPMDIsWorker(*pargv)) { /* we're an implicit worker using AMUDP SPMD API */
    int rank;
    AM_Safe(AMUDP_SPMDStartup(pargc, pargv, 0, 0, NULL, &gpid, &__sc_bundle, &__sc_endpoint));
    rank = AMUDP_SPMDMyProc();
    AM_Safe(AM_GetTranslationName(__sc_endpoint, rank, &__sc_endpoint_name)); /* lookup our name */
    usingAMUDPSPMD = 1; /* signal that we've completed endpoint initialization */
    if (getenvMaster("TI_BACKEND_SILENT")) {
      AMUDP_SilentMode = 1; /* suppress diagnostic output */
    }
  }
  else { /* decide whether we're an implicit master */
    int isExplicitWorker = 0;
    char spawnfn = '\0';
#ifdef REXEC 
    isExplicitWorker = (int)getenv("REXEC_PAR_DEGREE");
    spawnfn = 'R';
    { static char *AMUDP_DEFAULT_SPAWNFN_IDENT_STRING = "$TitaniumDefaultSpawnFunction: R $"; }
#elif defined(GLUNIX)
    isExplicitWorker = !Glib_AmIStartup();
    spawnfn = 'G';
    { static char *AMUDP_DEFAULT_SPAWNFN_IDENT_STRING = "$TitaniumDefaultSpawnFunction: G $"; }
#else /* AMUDP implicit ssh startup */
    isExplicitWorker = 0; /* we don't support explicit ssh startup */
    spawnfn = 'S';
    { static char *AMUDP_DEFAULT_SPAWNFN_IDENT_STRING = "$TitaniumDefaultSpawnFunction: S $"; }
#endif
    if (!isExplicitWorker) { /* we are an implicit master */
      spawnfn_t fp = NULL;
      int networkdepth = 0;
      int num_nodes = parseNumNodes(pargc, pargv);

      if (getenv("TI_SPAWNFN")) /* user has overridden default implicit job spawner */
        spawnfn = *getenv("TI_SPAWNFN");
      if (getenv("TI_UDPNETWORKDEPTH")) 
        networkdepth = atoi(getenv("TI_UDPNETWORKDEPTH"));
      if (networkdepth < 1) networkdepth = 20; /* a reasonable default for many systems */

      switch (spawnfn) {
#ifdef GLUNIX
        case 'g': case 'G':
          fp = AMUDP_SPMDGlunixSpawn;
          break;
#endif
        case 's': case 'S':
          fp = AMUDP_SPMDSshSpawn;
          break;
#ifdef REXEC
        case 'r': case 'R':
          fp = AMUDP_SPMDRexecSpawn;
          break;
#endif
        case 'l': case 'L':
          fp = AMUDP_SPMDLocalSpawn;
          break;
        default:
          fprintf (stderr, "Tic: Invalid spawn function specified in TI_SPAWNFN\n");
          fprintf (stderr, "Tic: The following mechanisms are available:\n");
          fprintf (stderr, 
            "    'S'  Spawn jobs using ssh remote shells\n"
            "    'L'  Spawn jobs using fork()/exec() on the local machine (good for SMP's)\n"
          #ifdef REXEC
            "    'R'  Spawn jobs using rexec (Berkeley Millennium only)\n"
          #endif
          #ifdef GLUNIX
            "    'G'  Spawn jobs using GLUNIX (Berkeley NOW only)\n"
          #endif
            );
          exit (1);
        }
    AM_Safe(AMUDP_SPMDStartup(pargc, pargv, num_nodes, networkdepth, fp, &gpid, NULL, NULL));
    abort(); /* master startup should never return */
    }
  }
#else /* not COMM_AMUDP */
#ifdef GLUNIX
  if (Glib_AmIStartup()) {
    int num_nodes = parseNumNodes(pargc, pargv);
      /* Spawn only returns if it failed */
    if (Glib_Spawn(num_nodes, (*pargv)[0], *pargv)) {
      perror("Global spawn failed");
      exit(-1);
    }
    abort();
  }
#endif /* Glunix */
#endif
}

void
free_resources ()
{
#ifdef COMM_AM2 
  if (usingAMUDPSPMD) {
#if defined(COMM_AMUDP) 
    AMUDP_SPMDExit(0);
#endif
    }
  else if (COMM_GetBParallelDegree() > 1) {
    AM_Safe(AM_FreeEndpoint (__sc_endpoint));
    AM_Safe(AM_FreeBundle (__sc_bundle));
    AM_Safe(AM_Terminate());
#ifdef HAVE_ENAMESERVER_H
    NS_Safe((NameServer_Finish(),0)); /* NameServer_Finish() has a void return */
#endif
  }
#endif /* AMII */
}

void
free_resource_handler (int sig)
{
    free_resources ();
#ifdef GLUNIX
  /* Glunix needs this to prevent orphaned processes */
  Glib_Signal(Glib_GetMyNpid(), VNN_ALL, SIGTERM);
#endif
    exit (0);
}

__tic_thread_key_t tic_thread_key;



int ___PROCS;
int ___BOXES;
Box ___MYBOX;

#ifdef MEMORY_SHARED
int ___MYBOXPROCS;
#else
Process ___MYPROC;
#endif

void memory_init();

void init_box_consts() {
  /* This process is a node in a parallel job. */
  ___PROCS = COMM_GetParallelDegree();
  ___BOXES = COMM_GetBParallelDegree();
  ___MYBOX = COMM_GetMyBoxNumber();
  #ifdef MEMORY_SHARED
    ___MYBOXPROCS = COMM_GetMyBParallelDegree();
  #else
    ___MYPROC = myProcessNumber();
  #endif

  com_init();
  store_init();
  memory_init();
  __tic_lock_init(&AM_lock);
}

char *RUNTIME_MODEL = TIC_BACKEND_NAME; 

/* these strings are read from the executable by ident and tcrun */
char *IDENT_STRING_1 = "$TitaniumBackend: " TIC_BACKEND_NAME " $"; 

extern void region_init(void);

void *tic_main(int argc, char **argv)
{
  ti_main(argc, argv);
  return 0;
}


void printGlobalStats();

void *thread_go(void *procinfo) {
  processInfo *procInfo = (processInfo *)procinfo;
  int result = __tic_thread_set_key_value(tic_thread_key, (void *)procInfo);
  if (result != 0) {
    printf("error in tic thread set key value: %d\n", result);
    return NULL;
  }

  #ifdef _CRAY
    /* Setup a trap handler for SIGFPE (floating point exception) to keep it from crashing us */
    sigignore(SIGFPE);
  #endif

  #ifdef DEBUG_THREADS
    fprintf(stderr,"Processor starting MYPROC=%i PROCS=%i MYBOX=%i BOXES=%i MYBOXPROC=%i MYBOXPROCS=%i\n", 
      MYPROC, PROCS, MYBOX, BOXES, MYBOXPROC, MYBOXPROCS); 
    fflush(stderr);
  #endif

  local_barrier(); /* wait for all local threads */

  region_init();
 
  monitor_startup();
 
  barrier();

  ti_main (procInfo->box->argc, procInfo->box->argv);

  barrier();

#ifdef COMM_AMUDP
  /* Statistical collection and reporting for AMUDP backends */
  if (getenvMaster("TI_AMSTATS") &&
      COMM_GetBParallelDegree() > 1) printGlobalStats();
#endif

  return NULL;
}

void spawn_threads() {
  int i;
  processInfo *thisproc;
  
  __tic_thread_key_init(&tic_thread_key);

  barrier_init(); /* this must happen on the primordial thread before thread spawn */

#ifdef MEMORY_DISTRIBUTED
  /* need an external barrier here to allow titanium barrier to be bootstrapped on all nodes */
#ifdef COMM_AMUDP
  if (usingAMUDPSPMD) AMUDP_SPMDBarrier(); 
  else 
#endif
  {
#if defined(GLUNIX)
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#elif defined( COMM_AMLAPI )
  lapi_handle_t *lapi_hndl = AMLAPI_GetLAPIHandle ();	/* Get My LAPI endpoint */
  LAPI_Gfence(*lapi_hndl);	/* Global barrier to sync before starting */
#else
  abort(); /* we need something else here - barrier() is no good */
  barrier(); /* this is not strictly correct and may cause problems */
#endif /* Glunix */
    }
#endif

  /* this thread will eventually act as thread 0 for this box */
  thisproc = COMM_GetHisBoxProcess(0);
  __tic_thread_set_key_value(tic_thread_key, thisproc);
  thisproc->threadID = __tic_thread_self();

  init_box_consts(); /* called on proc 0 before other threads spawned */

  /* spawn other threads */
  for (i = 1; i < COMM_GetMyBParallelDegree(); i++) {
    int result;
    thisproc = COMM_GetHisBoxProcess(i);
    __tic_create_thread(thread_go, (void *)thisproc, &thisproc->threadID, result);
    if (result != 0) {
      perror("tic_create_thread");
      exit(-1);
    }
    #ifdef DEBUG_THREADS
      printf("created thread 0x%08x\n",(int) thisproc->threadID);
    #endif
  }
  
  thisproc = COMM_GetHisBoxProcess(0);
  thread_go(thisproc); /* run process 0 */
  return;
}

void join_threads() {
  int i;
  
    for (i = 1; i < COMM_GetMyBParallelDegree(); i++) {
      int result;
      processInfo *thisproc = COMM_GetHisBoxProcess(i);

      #ifdef DEBUG_THREADS
        printf("joining thread 0x%08x\n",(int) thisproc->threadID);
      #endif

      result = __tic_thread_join(thisproc->threadID, NULL);

      if (result != 0) {
	perror("tic_join_threads");
	exit(-1);
      }
    }
}



int main (int argc, char** argv, char** envp) {
    int countBoxMem, countProcMem;
    char *boxMem, *procMem;
    char	hostname[MAXHOSTNAMELEN] = "(unknown)";

    /* do this *very* early, in case library routines allocate memory */
    ti_mem_thr_init(&argv); /* init the GC for the main thread - not necessary for dynamically created threads, should only be run on the primordial thread */

    sigset (SIGHUP, free_resource_handler);
    sigset (SIGINT, free_resource_handler);
    sigset (SIGQUIT, free_resource_handler);
    sigset (SIGTERM, free_resource_handler);

    progName = argv[0];

    setup_global_layer(&argc, &argv); /* must come before any worker-related code, as this may be the startup master node */

    countBoxMem = COMM_Init_One();
    #if HAVE_ALLOCA_H 
      boxMem = alloca(countBoxMem);
    #else
      boxMem = malloc(countBoxMem);
    #endif

    gethostname(hostname, sizeof(hostname));
    
    if (getenvMaster("TI_FREEZE"))
    {
      volatile int frozen = 1;

      printf("frozen for debugger: host %s, pid %d\n", hostname, getpid());
      fflush(stdout);
      
      while (frozen) {
        #if HAVE_SCHED_YIELD
          sched_yield();
        #endif
        }
    }


    countProcMem = COMM_Init_Two(boxMem);
    #if HAVE_ALLOCA_H 
      procMem = alloca(countProcMem);
    #else
      procMem = malloc(countProcMem);
    #endif

    COMM_Init_Three(procMem);

#if defined(MEMORY_DISTRIBUTED) || defined(MEMORY_SHARED)
  if (!getenvMaster("TI_BACKEND_SILENT")) {
    printf("Tic: %s is box number %d of %d\n", hostname, 
	 COMM_GetMyBoxNumber(), COMM_GetBParallelDegree());
    fflush(stdout);
    }
#endif

    if (COMM_GetBParallelDegree() > 1) {
#ifdef COMM_AM2
      setupAM();
#else /* COMM_AM2 */
      fprintf(stderr, "Error: Trying to use more than one box without a known communications layer.\n");
      return 2;
#endif /* COMM_AM2 */
    }
    
    COMM_Init_Four(argc, argv);
    
    if (COMM_GetBParallelDegree() > 1) {
#ifdef COMM_AM2
      if (usingAMUDPSPMD) {
#ifdef COMM_AMUDP
        AM_Safe(AMUDP_SPMDBarrier()); /* need a barrier after handler setup */
#endif
        }
      else setupNames();
#ifdef COMM_AMUDP
    { int networkdepth = 0;
      if (getenv("TI_UDPNETWORKDEPTH")) 
        networkdepth = atoi(getenv("TI_UDPNETWORKDEPTH"));
      if (networkdepth < 1) networkdepth = 20; /* a reasonable default for many systems */
      /* AMUDP requires us to specify the number of peers and the desired network depth */
      if (!usingAMUDPSPMD) /* SPMD startup already takes care of this */
        AM_Safe(AM_SetExpectedResources(__sc_endpoint, COMM_GetBParallelDegree(), networkdepth));
    }
#endif
#else /* COMM_AM2 */
      fprintf(stderr, "Error: Trying to use more than one box without a known communications layer.\n");
      return 2;
#endif /* COMM_AM2 */
    }

    __tic_thread_set_concurrency(COMM_GetMyBParallelDegree());

    spawn_threads();
    
    join_threads();

    fflush(stdout);
    fflush(stderr);

#ifdef GLUNIX
    /* don't free AM endpoints until we're sure everyone is ready (AM-based barrier() not good enough) */
    if (!Glib_Barrier()) {
      perror("global barrier");
      exit(-1);
    }
#endif /* Glunix */

    free_resources ();
    
    return 0; /* if Titanium program returned from void main, that means return 0 */
}

/* Statistical collection and reporting for AMUDP backends */
#ifdef COMM_AMUDP
static int statscalls = 0;
static amudp_stats_t globalStats;

void ___amstats_request_handler(void *token, void *buf, int nbytes, int procnum, int reserved) {
  assert(nbytes == sizeof(amudp_stats_t));
  AM_Safe(AMUDP_AggregateStatistics(&globalStats, (amudp_stats_t *)buf));
  statscalls++;
  }
void printGlobalStats() {
  amudp_stats_t stats;
  statscalls = 0; 
  globalStats = AMUDP_initial_stats;

  /* make sure we're done sending msgs for now */
  barrier();

  if (MYBOXPROC == 0) {
    if (usingAMUDPSPMD) AMUDP_SPMDBarrier(); /* try to minimize impact of monitoring */

    AM_Safe(AMUDP_GetEndpointStatistics(CUR_ENDPOINT, &stats)); /* get statistics */

    /* don't let stats msgs interfere with display */
    if (usingAMUDPSPMD) AMUDP_SPMDBarrier(); 

    assert(sizeof(amudp_stats_t) < AM_MaxMedium());
    TI_AM_RequestI2(CUR_ENDPOINT, 0, __AMSTATS_H, &stats, sizeof(amudp_stats_t), MYBOX, 0); /* send to zero */

    if (MYPROC == 0) {
      while (statscalls < BOXES) {
        tic_poll(); 
        sched_yield();
        }

      fprintf(stderr, "--------------------------------------------------\n"
                      "Global AM2 usage statistics:\n");
      AM_Safe(AMUDP_DumpStatistics(stderr, &globalStats, 1));
      fprintf(stderr, "--------------------------------------------------\n");
      fflush(stderr);

      sleep(1); /* HACK: give a little time for this output to reach user's console before shutdown */
      }
    }

  barrier(); /* just to keep things clean */

  }
#else
void printGlobalStats() {
  }
#endif
