/*									tab:8
 *
 * bulk.c - Handlers for bulk functions.
 *
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Steve Luna
 * Version:		28
 * Creation Date:	Sat Nov 27 23:07:14 1993
 * Filename:		bulk.c
 * History:
 *	LTL	8	Sun Jul 10 15:06:21 1994
 *		Modified for TCPAM
 *	RPM	2	Sun Nov 28 22:33:34 1993
 *		new version for PMAM interface, bonged segments
 */

#include <tic.h>

#ifdef COMM_AM2

#ifdef PTR64

void ___get_handler(void *token, juint addr0, juint addr1, 
		    juint retAddr0, juint retAddr1,  
		    juint len0, juint len1,
		    juint ctr0, juint ctr1) { 
  void *addr; 
  void *retAddr;
  julong len;
  Counter *ctr; 

  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&retAddr, retAddr0); 
  SETLOWWORD((julong *)&retAddr, retAddr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __get_handler(token, addr, retAddr, len, ctr); \
} 

void __get_handler(void *token, void *local, void *dest, julong len, Counter *ctr) {
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
  TI_AM_ReplyXfer1(token, (jUIntPointer)dest, __GETCOMPLETE_RH, (jUIntPointer)local,
		   len, (jUIntPointer)ctr);
#else
  TI_AM_ReplyI2(token, __GETCOMPLETE_RH, (jUIntPointer)local,
		   len, (jUIntPointer)dest, (jUIntPointer)ctr);
#endif
}

#if AM2_HAS_HUGE_SEGMENTS 
void ___get_complete_reply_handler (void *token, juint addr0, juint addr1, 
				    juint len0, juint len1, juint ctr0, juint ctr1) {
  void *addr; 
  julong len;
  Counter *ctr; 

  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __get_complete_reply_handler(token, (void *)addr, len, (Counter *)ctr);
}

void __get_complete_reply_handler (void *token, void *buf, julong len, Counter *ctr)
{
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}
#else
void ___get_complete_reply_handler (void *token, juint addr0, juint addr1, 
				    juint len0, juint len1, juint da0, juint da1,
                                    juint ctr0, juint ctr1) {
  void *addr; 
  julong len;
  void *destaddr;
  Counter *ctr; 

  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&destaddr, da0); 
  SETLOWWORD((julong *)&destaddr, da1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __get_complete_reply_handler(token, (void *)addr, len, destaddr, (Counter *)ctr);
}

void __get_complete_reply_handler (void *token, void *buf, julong len, void *destaddr, Counter *ctr)
{
  memcpy(destaddr, buf, len);
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}
#endif

#else

void ___get_handler(void *token, juint addr,  
		    juint retAddr, juint len, 
		    juint ctr) {  
  __get_handler(token, (void *)addr, (void *)retAddr, len, (Counter *)ctr); \
}

void __get_handler(void *token, void *local, void *dest, juint len, Counter *ctr) {
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
  TI_AM_ReplyXfer1(token, (jUIntPointer)dest, __GETCOMPLETE_RH, (jUIntPointer)local,
		   len, (jUIntPointer)ctr);
#else
  TI_AM_ReplyI2(token, __GETCOMPLETE_RH, local,
		   len, (jUIntPointer)dest, (jUIntPointer)ctr);
#endif
}

#if AM2_HAS_HUGE_SEGMENTS 
void ___get_complete_reply_handler (void *token, juint addr, juint len, juint ctr) {
  __get_complete_reply_handler(token, (void *)addr, len, (Counter *)ctr);
}

void __get_complete_reply_handler (void *token, void *buf, juint len, Counter *ctr)
{
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}
#else
void ___get_complete_reply_handler (void *token, juint addr, juint len, juint destaddr, juint ctr) {
  __get_complete_reply_handler(token, (void *)addr, len, (void *)destaddr, (Counter *)ctr);
}

void __get_complete_reply_handler (void *token, void *buf, juint len, void *destaddr, Counter *ctr)
{ memcpy(destaddr, buf, len);
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}
#endif

#endif  /* PTR64 */

				  


#ifdef PTR64
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
void ___bulk_put_handler(void *token,  
			 juint addr0, juint addr1, 
			 juint len0, juint len1,  
			 juint ctr0, juint ctr1) { 
  void *addr; 
  julong len;
  Counter *ctr; 
  
  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 
  
  __bulk_put_handler(token, addr, len, ctr); \
}

void __bulk_put_handler(void *token, void *buf, julong len, Counter *ctr)
{
  TI_AM_Reply1(token, __PUTCTRCOMPSRC_RH, (jUIntPointer)ctr);
}
#else
void ___bulk_put_handler(void *token,  
			 juint addr0, juint addr1, 
			 juint len0, juint len1, 
                         juint da0, juint da1,
			 juint ctr0, juint ctr1) { 
  void *addr; 
  julong len;
  Counter *ctr; 
  void *destaddr; 
  
  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 
  SETHIGHWORD((julong *)&destaddr, da0); 
  SETLOWWORD((julong *)&destaddr, da1); 
  
  __bulk_put_handler(token, addr, len, destaddr, ctr); \
}

void __bulk_put_handler(void *token, void *buf, julong len, void *destaddr, Counter *ctr)
{
  memcpy(destaddr, buf, len);
  TI_AM_Reply1(token, __PUTCTRCOMPSRC_RH, (jUIntPointer)ctr);
}
#endif

void ___put_ctr_comp_src_reply_handler(void *token, juint ctr0, juint ctr1) {
{
  Counter *ctr; 

  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __put_ctr_comp_src_reply_handler(token, (Counter *)ctr);
}

void __put_ctr_comp_src_reply_handler(void *token, Counter *ctr)
{
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}

#else
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
void ___bulk_put_handler(void *token, juint addr,  
			 juint len, juint ctr) {  
  __bulk_put_handler(token, (void *)addr, len, (Counter *)ctr);
}

void __bulk_put_handler(void *token, void *buf, juint len, Counter *ctr)
{
  TI_AM_Reply1(token, __PUTCTRCOMPSRC_RH, (jUIntPointer)ctr);
}
#else
void ___bulk_put_handler(void *token, juint addr,  
			 juint len, juint destaddr, juint ctr) {  
  __bulk_put_handler(token, (void *)addr, len, (void *)destaddr, (Counter *)ctr);
}

void __bulk_put_handler(void *token, void *buf, juint len, void *destaddr, Counter *ctr)
{
  memcpy(destaddr, buf, len);
  TI_AM_Reply1(token, __PUTCTRCOMPSRC_RH, (jUIntPointer)ctr);
}
#endif

void ___put_ctr_comp_src_reply_handler(void *token, juint ctr) {
  __put_ctr_comp_src_reply_handler(token, (Counter *)ctr);
}

void __put_ctr_comp_src_reply_handler(void *token, Counter *ctr) {
  if (ctr != NULL) {
    __decr_ctr(ctr);
  }
}

#endif /* PTR64 */


void __null_reply_handler(void *token)
{
}

#endif /* AMII */

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */


#ifdef COMM_AM2

#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */

#ifdef PTR64

void ___bulk_store_handler(void *token, juint addr0, juint addr1, juint len0, juint len1, juint ctr0, juint ctr1) {
  
  void *addr; 
  julong len;
  Counter *ctr;
  
  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __bulk_store_handler(token, addr, len, ctr);
}

void __bulk_store_handler(void *token, void *buf, julong len, Counter *ctr) {
  __incr_ctr_size(ctr, len);
  TI_AM_Reply0(token, __NULL_RH);
}

#else

void ___bulk_store_handler(void *token, juint addr, juint len, juint ctr) {
  __bulk_store_handler(token, (void *)addr, len, (Counter *)ctr);
}

void __bulk_store_handler(void *token, void *buf, juint len, Counter *ctr) {
  __incr_ctr_size(ctr, len);
  TI_AM_Reply0(token, __NULL_RH);
}

#endif /* PTR64 */

#else

#ifdef PTR64

void ___bulk_store_handler(void *token, juint addr0, juint addr1, juint len0, juint len1, 
                           juint da0, juint da1, juint ctr0, juint ctr1) {
  
  void *addr; 
  julong len;
  void *destaddr;
  Counter *ctr;
  
  SETHIGHWORD((julong *)&addr, addr0); 
  SETLOWWORD((julong *)&addr, addr1); 
  SETHIGHWORD((julong *)&len, len0); 
  SETLOWWORD((julong *)&len, len1); 
  SETHIGHWORD((julong *)&destaddr, da0); 
  SETLOWWORD((julong *)&destaddr, da1); 
  SETHIGHWORD((julong *)&ctr, ctr0); 
  SETLOWWORD((julong *)&ctr, ctr1); 

  __bulk_store_handler(token, addr, len, ctr);
}

void __bulk_store_handler(void *token, void *buf, julong len, void *destaddr, Counter *ctr) {
  __incr_ctr_size(ctr, len);
  memcpy(destaddr, buf, len);
  TI_AM_Reply0(token, __NULL_RH);
}

#else

void ___bulk_store_handler(void *token, juint addr, juint len, juint destaddr, juint ctr) {
  __bulk_store_handler(token, (void *)addr, len, (void *)destaddr, (Counter *)ctr);
}

void __bulk_store_handler(void *token, void *buf, juint len, void *destaddr, Counter *ctr) {
  __incr_ctr_size(ctr, len);
  memcpy(destaddr, buf, len);
  TI_AM_Reply0(token, __NULL_RH);
}

#endif /* PTR64 */

#endif /* HUGE_SEGMENTS */

#endif /* AMII */



/* STORE bulk */
void __store(jGPointer gPtr, void *lPtr, int len) {
  __bulk_store_ctr(gPtr, lPtr, len, &recvBytes, &sentBytes);
}

void __bulk_store_ctr(jGPointer gPtr, void *lPtr, int len, Counter *rbytes, Counter *sbytes)
{
  int box       = tobox(gPtr);
  void *addr    = tolocal(gPtr);

  if(box == MYBOX) {
    __incr_ctr_size(sbytes, len);
    memcpy(addr,lPtr, len);
    __incr_ctr_size(rbytes, len);
  } 
#ifdef COMM_AM2
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
  else {
    int i;
    if (len <= ___AM_MAX_LONG) {
      ADVANCE_SEND_ENDPOINT;
      TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET, 
			 __BULK_STORE_H, (void *)lPtr, len, (jUIntPointer)rbytes);
    }
    else {
      for (i = ___AM_MAX_LONG; i < len; i += ___AM_MAX_LONG) {
	ADVANCE_SEND_ENDPOINT;
	TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET, 
			   __BULK_STORE_H, (void *)lPtr, ___AM_MAX_LONG, (jUIntPointer)rbytes); 
	(*(char **)&addr) += ___AM_MAX_LONG;
	(*(char **)&lPtr) += ___AM_MAX_LONG;
      }
      i -= ___AM_MAX_LONG;
      ADVANCE_SEND_ENDPOINT;
      TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET, 
			 __BULK_STORE_H, (void *)lPtr, len - i, (jUIntPointer)rbytes);
    }
    
    __incr_ctr_size(sbytes, len);
  }
#else
  else {
    int i;
    if (len <= ___AM_MAX_MEDIUM) {
      ADVANCE_SEND_ENDPOINT;
      TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_STORE_H, 
        (void *)lPtr, len, addr, (jUIntPointer)rbytes);
    }
    else {
      for (i = ___AM_MAX_MEDIUM; i < len; i += ___AM_MAX_MEDIUM) {
	ADVANCE_SEND_ENDPOINT;
	TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_STORE_H, 
          (void *)lPtr, ___AM_MAX_MEDIUM, addr, (jUIntPointer)rbytes); 
	(*(char **)&addr) += ___AM_MAX_MEDIUM;
	(*(char **)&lPtr) += ___AM_MAX_MEDIUM;
      }
      i -= ___AM_MAX_MEDIUM;
      ADVANCE_SEND_ENDPOINT;
      TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_STORE_H, 
        (void *)lPtr, len - i, addr, (jUIntPointer)rbytes); 
    }
    
    __incr_ctr_size(sbytes, len);
  }
#endif
#endif /* AMII */
}
