/* 									tab:8
 *
 * bulk.h - Global memory bulk access.
 *          Implementation dependent definitions and declarations.
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:		Andrea Dusseau
 * Version:		46
 * Creation Date:	Thu Oct  8 14:16:42 PDT 1992
 * Filename:		bulk.h
 * History:
 *	LTL	16	Sun Jul 10 15:05:54 1994
 *		Modified for TCPAM
 *	ssl	9	Sat Nov 27 22:58:42 1993
 *		Included __get* __put* __write from memory.[ch]. Included split-c.h
 *	ssl	8	Tue Nov 16 11:56:20 1993
 *		Replaced global tags with macros.
 *	AK	7	Mon Apr 12 15:13:03 1993
 *		Changed bulk_store decl to "inline". To be consistent
 *		with split-c.h
 *	AK	4	Sat Jan 23 02:38:09 1993
 *		Made bulk_store independant of __store.
 * 	TvE	2	Thu Dec 24 04:30:31 PST 1992
 *		Rewrite. Now entirely based on PUTs, GETs and STOREs
 *	AK	1	Sat Nov 28 21:17:06 1992
 *		Changed include file name from cmam2-4
 *	DEC	1	Sat Oct 24 01:21:37 1992
 *		Changed store counters to recvBytes and sendBytes to be
 *      	compatible with split-c_int of non-bulk ops.
 */

#ifndef __BULK_INT_H__
#define __BULK_INT_H__

#include <string.h> /* memcpy */

#include <tic.h>

/* ....................
 *
 * Bulk PUTs and WRITEs
 *
 * ....................
 */



/* .........
 *
 * Bulk GETs
 *
 * .........
 */

/* Explicit counter */
EXTERN INLINE_MODIFIER void bulk_get_ctr(void *lptr, jGPointer gptr,
				jint len, Counter *ctr)
INLINE({
	__get_ctr(lptr, gptr, len, ctr);
})

/* Implicit GET counter */
EXTERN INLINE_MODIFIER void bulk_get(void *lptr, jGPointer gptr, jint len)
INLINE({
	__get_ctr(lptr, gptr, len, &numGetOuts);
})

/* Blocking */
EXTERN INLINE_MODIFIER void bulk_read(void *lptr, jGPointer gptr, jint len)
INLINE({
	Counter ctr = 0;
	__get_ctr(lptr, gptr, len, &ctr);
	sync_ctr(&ctr);
})

/* GET bulk with explicit counter */
EXTERN INLINE_MODIFIER void __get_ctr(void *lPtr, jGPointer gPtr,
				      jint len, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
     INLINE({
       int box    = tobox(gPtr);
       void *addr = tolocal(gPtr);
       int i;

       if (len == 0) return; /* this actually happens sometimes */

       if (box == MYBOX) {
	 memcpy(lPtr,addr, len);
       } 
       else {
	 if (len <= ___AM_MAX_LONG) {
	   __incr_ctr(ctr);
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			  (jUIntPointer)lPtr, len, (jUIntPointer)ctr);
	 }
	 else {
	   __add_ctr(ctr, (len + ___AM_MAX_LONG - 1) / ___AM_MAX_LONG);
	   for (i = ___AM_MAX_LONG; i < len; i += ___AM_MAX_LONG) {
	     ADVANCE_SEND_ENDPOINT;
	     TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			    (jUIntPointer)lPtr, ___AM_MAX_LONG, 
			    (jUIntPointer)ctr);
	     (*(char **)&addr) += ___AM_MAX_LONG;
	     (*(char **)&lPtr) += ___AM_MAX_LONG;
	   }
	   i -= ___AM_MAX_LONG;
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			  (jUIntPointer)lPtr, len - i,  (jUIntPointer)ctr);
	 }
       }
     })
#else
     INLINE({
       int box    = tobox(gPtr);
       void *addr = tolocal(gPtr);
       int i;

       if (len == 0) return; /* this actually happens sometimes */

       if (box == MYBOX) {
	 memcpy(lPtr,addr, len);
       } 
       else {
	 if (len <= ___AM_MAX_MEDIUM) {
	   __incr_ctr(ctr);
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			  (jUIntPointer)lPtr, len, (jUIntPointer)ctr);
	 }
	 else {
	   __add_ctr(ctr, (len + ___AM_MAX_MEDIUM - 1) / ___AM_MAX_MEDIUM);
	   for (i = ___AM_MAX_MEDIUM; i < len; i += ___AM_MAX_MEDIUM) {
	     ADVANCE_SEND_ENDPOINT;
	     TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			    (jUIntPointer)lPtr, ___AM_MAX_MEDIUM, 
			    (jUIntPointer)ctr);
	     (*(char **)&addr) += ___AM_MAX_MEDIUM;
	     (*(char **)&lPtr) += ___AM_MAX_MEDIUM;
	   }
	   i -= ___AM_MAX_MEDIUM;
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_Request4(CUR_ENDPOINT, box, __GET_H, (jUIntPointer)addr, 
			  (jUIntPointer)lPtr, len - i,  (jUIntPointer)ctr);
	 }
       }
     })
#endif
#else
     INLINE({
       void *addr = tolocal(gPtr);
       memcpy(lPtr,addr, len);
     })
#endif /* COMM_AM2 */


EXTERN INLINE_MODIFIER void __get(void *lPtr, jGPointer gPtr, int len)
INLINE({ __get_ctr(lPtr, gPtr, len, &numGetOuts); })
	

/* .........
 *
 * Bulk PUTs
 *
 * .........
 */

EXTERN INLINE_MODIFIER void bulk_put_ctr(jGPointer gptr, void *lptr,
				jint len, Counter *ctr)
     INLINE({
       __put_ctr(gptr, lptr, len, ctr);
     })

EXTERN INLINE_MODIFIER void bulk_put(jGPointer gptr, void *lptr, jint len)
     INLINE({
       __put(gptr, lptr, len);
     })

EXTERN INLINE_MODIFIER void bulk_write(jGPointer gptr, void *lptr, jint len)
     INLINE({
       __ticwrite(gptr, lptr, len);
     })

extern char *___AM_SEGOFFSET;

EXTERN INLINE_MODIFIER void __put_ctr(jGPointer gPtr, void *lPtr, int len, Counter *ctr __attribute__((unused)))
#ifdef COMM_AM2
#if AM2_HAS_HUGE_SEGMENTS /* AM_MaxSeg is currently too small (on now & millenium) for this to work */
     INLINE({
       int box       = tobox(gPtr);
       void *addr    = tolocal(gPtr);
       int i;

       if (len == 0) return; /* this actually happens sometimes */

       if(box == MYBOX) {
	 memcpy(addr,lPtr, len);
       } 
       else {
	 if (len <= ___AM_MAX_LONG) {
	   __incr_ctr(ctr);
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET,
			      __BULK_PUT_H, (void *)lPtr,
			      len, (jUIntPointer)ctr);
	 }
	 else {
	   __add_ctr(ctr, (len + ___AM_MAX_LONG - 1) / ___AM_MAX_LONG);
	   for (i = ___AM_MAX_LONG; i < len; i += ___AM_MAX_LONG) {
	     ADVANCE_SEND_ENDPOINT;
	     TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET, 
				__BULK_PUT_H, (void *)lPtr,
				___AM_MAX_LONG, (jUIntPointer)ctr);
	     (*(char **)&addr) += ___AM_MAX_LONG;
	     (*(char **)&lPtr) += ___AM_MAX_LONG;
	   }
	   i -= ___AM_MAX_LONG;
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_RequestXfer1(CUR_ENDPOINT, box, ((char*)addr)-___AM_SEGOFFSET, 
			      __BULK_PUT_H, (void *)lPtr,
			      len - i, (jUIntPointer)ctr);
	 }
       }
     })
#else
     INLINE({
       int box       = tobox(gPtr);
       void *addr    = tolocal(gPtr);
       int i;

       if (len == 0) return; /* this actually happens sometimes */

       if(box == MYBOX) {
	 memcpy(addr,lPtr, len);
       } 
       else {
	 if (len <= ___AM_MAX_MEDIUM) {
	   __incr_ctr(ctr);
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_PUT_H, 
             (void *)lPtr, len, addr, (jUIntPointer)ctr);
	 }
	 else {
	   __add_ctr(ctr, (len + ___AM_MAX_MEDIUM - 1) / ___AM_MAX_MEDIUM);
	   for (i = ___AM_MAX_MEDIUM; i < len; i += ___AM_MAX_MEDIUM) {
	     ADVANCE_SEND_ENDPOINT;
	     TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_PUT_H, 
               (void *)lPtr, ___AM_MAX_MEDIUM, addr, (jUIntPointer)ctr);
	     (*(char **)&addr) += ___AM_MAX_MEDIUM;
	     (*(char **)&lPtr) += ___AM_MAX_MEDIUM;
	   }
	   i -= ___AM_MAX_MEDIUM;
	   ADVANCE_SEND_ENDPOINT;
	   TI_AM_RequestI2(CUR_ENDPOINT, box, __BULK_PUT_H, 
             (void *)lPtr, len - i, addr, (jUIntPointer)ctr);
	 }
       }
     })
#endif
#else
     INLINE({
       void *addr    = tolocal(gPtr);
       memcpy(addr,lPtr, len);
     })
#endif /* COMM_AM2 */     
     

EXTERN INLINE_MODIFIER void __put(jGPointer gPtr, void *lPtr, int len)
     INLINE({
       __put_ctr(gPtr,lPtr,len,&numPutOuts);
     })
     
EXTERN INLINE_MODIFIER void __ticwrite(jGPointer gPtr, void *lPtr, int len)
     INLINE({
       __put_ctr(gPtr,lPtr,len,&numPutOuts);
       sync_ctr(&numPutOuts);
     })


/* ...........
 *
 * Bulk STOREs
 *
 * ...........
 */


EXTERN void __bulk_store_ctr(jGPointer gptr, void *lptr, int len, Counter *rbytes, Counter *sbytes);
extern void __null_reply_handler(void *token);


#ifdef PTR64
extern void ___get_handler(void *token, juint addr0, juint addr1, 
			   juint retAddr0, juint retAddr1, 
			   juint len0, juint len1, 
			   juint ctr0, juint ctr1);


extern void ___put_ctr_comp_src_reply_handler(void *token, juint ctr0, juint ctr1);

#ifdef AM2_HAS_HUGE_SEGMENTS
extern void ___bulk_store_handler(void *token, juint addr0, juint addr1,
			    juint len0, juint len1, juint ctr0, juint ctr1);
extern void __bulk_store_handler(void *token, void *buf, julong len, Counter *ctr);
extern void ___bulk_put_handler(void *token,  
				juint addr0, juint addr1, 
				juint len0, juint len1, 
				juint ctr0, juint ctr1);
extern void __bulk_put_handler(void *token, void *buf, julong len, Counter *ctr);
extern void ___get_complete_reply_handler(void *token, juint addr0,
					  juint addr1, juint len0, juint len1,
					  juint ctr0, juint ctr1);
extern void __get_complete_reply_handler(void *token, void *buf, julong len, Counter *ctr);
#else
extern void ___bulk_store_handler(void *token, juint addr0, juint addr1, juint len0, juint len1, 
                           juint da0, juint da1, juint ctr0, juint ctr1);
extern void __bulk_store_handler(void *token, void *buf, julong len, void *destaddr, Counter *ctr);
extern void ___bulk_put_handler(void *token,  
				juint addr0, juint addr1, 
				juint len0, juint len1, 
				juint da0, juint da1, juint ctr0, juint ctr1);
extern void __bulk_put_handler(void *token, void *buf, julong len, void *destaddr, Counter *ctr);
extern void ___get_complete_reply_handler(void *token, juint addr0,
					  juint addr1, juint len0, juint len1,
                                          juint da0, juint da1,
					  juint ctr0, juint ctr1);
extern void __get_complete_reply_handler(void *token, void *buf, julong len, void *destaddr, Counter *ctr);
#endif

extern void __get_handler(void *token, void *localAddr, void *destAddr,
			  julong len, Counter *ctr);

extern void __put_ctr_comp_src_reply_handler(void *token, Counter *ctr);



#else
extern void ___get_handler(void *token, juint addr, juint retAddr, juint len, juint ctr);



extern void ___put_ctr_comp_src_reply_handler(void *token, juint ctr);

#ifdef AM2_HAS_HUGE_SEGMENTS
extern void ___bulk_store_handler(void *token, juint addr, juint len, juint ctr);
extern void __bulk_store_handler(void *token, void *buf, juint len, Counter *ctr);
extern void ___bulk_put_handler(void *token, juint addr, juint len, juint ctr); 
extern void __bulk_put_handler(void *token, void *buf, juint len, Counter *ctr);
extern void ___get_complete_reply_handler(void *token, juint addr, juint len, juint ctr);
extern void __get_complete_reply_handler(void *token, void *buf, juint len, Counter *ctr);
#else
extern void ___bulk_store_handler(void *token, juint addr, juint len, juint destaddr, juint ctr);
extern void __bulk_store_handler(void *token, void *buf, juint len, void *destaddr, Counter *ctr);
extern void ___bulk_put_handler(void *token, juint addr, juint len, juint destaddr, juint ctr); 
extern void __bulk_put_handler(void *token, void *buf, juint len, void *destaddr, Counter *ctr);
extern void ___get_complete_reply_handler(void *token, juint addr, juint len, juint destaddr, juint ctr);
extern void __get_complete_reply_handler(void *token, void *buf, juint len, void *destaddr, Counter *ctr);
#endif


extern void __get_handler(void *token, void *localAddr, void *destAddr,
			  juint len, Counter *ctr);
extern void __put_ctr_comp_src_reply_handler(void *token, Counter *ctr);



#endif  /* PTR64 */



#endif








