/*									tab:8
 *
 * store.c - signaling global memory accesses (store)
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Thorsten von Eicken
 * Version:		71
 * Creation Date:	Thu Dec 24 02:39:52 PST 1992
 * Filename:		store.sc
 * History:
 *	ssl	61	Sun Feb 27 16:07:30 1994
 *		added all_store_sync_ctr
 *	ssl	60	Tue Feb  8 16:35:47 1994
 *		Added barrier to end of all_store_sync. Race condition if ___i_reduce
 *      is synthesized or if interrupts are turned on.
 *	ssl	59	Tue Dec 14 01:19:45 1993
 *		Moved store code to bulk.c
 *	ssl	58	Sun Dec 12 23:17:20 1993
 *		All bulk store stuff, except for simple case, removed for paragon.
 *	ssl	57	Sun Dec  5 23:07:07 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	55	Tue Nov 16 12:22:50 1993
 *		Replaced global tags with macros.
 *	LTL	53	Mon Aug 16 15:56:37 1993
 *		Fixed it up for version 112 to support CMMD
 *	AK	49	Tue Mar 16 14:31:28 1993
 *		Made use of both networks for store operations.
 *	AK	37	Fri Jan 22 21:23:08 1993
 *		Merged bulk.sc and removed it.
 *	AK	35	Fri Jan 22 21:20:17 1993
 *		Made __3i_store_handler public. Hope to merge
 *		bulk.sc in this file and make it static again.
 * 	TvE	33	Sun Jan 17 02:50:37 PST 1993
 *		___store with a short length had bugs. Didn't increment
 *		all pointers and didn't take odd lengths into account. Still
 *		needs a fix if the start address isn't on a word boundary.
 * 	TvE	1	Thu Dec 24 02:39:52 PST 1992
 *		Extracted from tic_int.h
 */

#include <tic.h>

__tic_lock_t	Store_mutex;

void store_init(void)
{
  if (MYBOXPROC == 0) {
    __tic_lock_init(&Store_mutex);
  }
}

/* .....................
 *
 * Keep track of STORESs
 *
 * .....................
 */

Counter sentBytes = 0;
Counter recvBytes = 0;

#ifdef COMM_AM2
/* Wait until all stores from all processors have completed */
void all_store_sync(void)
{
	int res;

	barrier();
	do {
		res = all_reduce_to_all_iadd(recvBytes-sentBytes);
	} while(res != 0);

	__tic_lock(&Store_mutex);
	recvBytes = sentBytes = 0;
	__tic_unlock(&Store_mutex);
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	barrier();
}
#else
#define all_store_sync()
#endif /* AMII */

#ifdef COMM_AM2
void all_store_sync_ctr(Counter *rbytes, Counter *sbytes)
{
	int res;

	barrier();
	do {
		res = all_reduce_to_all_iadd(*rbytes - *sbytes);
	} while(res != 0);

	__tic_lock(&Store_mutex);
	*rbytes = *sbytes = 0;
	__tic_unlock(&Store_mutex);
	/* need this if ___i_reduce is synthesized or if interrupts are turned
	   on - ssl */
	barrier();
}
#else 
#define all_store_sync_ctr(a, b)
#endif /* AMII */

/* ............................
 *
 * STOREs with explicit counter
 *
 * ............................
 */


#ifdef COMM_AM2

___STORE_HANDLER4(b, jbyte)
___STORE_HANDLER4(sh, jshort)
___STORE_HANDLER4(i, jint)
___STORE_HANDLER4(f, jfloat)
___STORE_HANDLER8(d, jdouble)
___STORE_HANDLER8(l, jlong)
#ifdef PTR64
     ___STORE_HANDLER16(gp, jGPointer, jUIntPointer)
     ___STORE_HANDLER8(lp, void *)
#else
void ___gp_store_handler(void *token, juint addr, juint box, juint proc, juint addrData,
			    juint ctr) {
  jGPointer data;
  mutateglobalwithbox(data, proc, box, addrData);
  __gp_store_handler(token, (jGPointer *)addr, data, (Counter *)ctr);
}
     ___STORE_HANDLER4(lp, void *)
#endif

#endif /* COMM_AM2 */

#ifdef COMM_AM2
void __b_store_handler (void *token, jbyte *addr, jbyte data, Counter *ctr)
{
  addr[0] = data;
  __incr_ctr_size(ctr, sizeof(jbyte));
  TI_AM_Reply0(token, __NULL_RH);
}

void __sh_store_handler(void *token, jshort *addr, jshort data, Counter *ctr)
{
	addr[0] = data;
	__incr_ctr_size(ctr, sizeof(jshort));

	TI_AM_Reply0(token, __NULL_RH);
}

void __i_store_handler(void *token, jint *addr, jint data, Counter *ctr)
{
	addr[0] = data;
	__incr_ctr_size(ctr, sizeof(jint));

	TI_AM_Reply0(token, __NULL_RH);
}

void __f_store_handler(void *token, jfloat *addr, jfloat data, Counter *ctr)
{
  addr[0] = data;
  __incr_ctr_size(ctr, sizeof(jfloat));
  
  TI_AM_Reply0(token, __NULL_RH);
}


void __d_store_handler(void *token, jdouble *addr, jdouble data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jdouble));
  TI_AM_Reply0(token, __NULL_RH);
}

void __l_store_handler(void *token, jlong *addr, jlong data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jlong));
  TI_AM_Reply0(token, __NULL_RH);
}


void __lp_store_handler(void *token, void **addr, void *data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(void *));
  TI_AM_Reply0(token, __NULL_RH);
}


#ifdef PTR64
void __gp_store_handler(void *token, jGPointer *addr, jUIntPointer data0,
		      jUIntPointer data1, Counter *ctr)
{
  *addr = mutateglobalwithbox(*addr, boxproctoproc(data0), 
			      boxproctobox(data0), (void *)data1);
  __incr_ctr_size(ctr, sizeof(jGPointer));
  TI_AM_Reply0(token, __NULL_RH);
}
#else
void __gp_store_handler(void *token, jGPointer *addr,
			jGPointer data, Counter *ctr)
{
  *addr = data;
  __incr_ctr_size(ctr, sizeof(jGPointer));
  TI_AM_Reply0(token, __NULL_RH);
}
#endif



#endif /* AMII */

