/* Code to support distributed garbage collection. Symlink this file into
   every backend that needs it. */

/* Right now there is no DGC...we have local GCs that conservatively 
   assume that pointers that escape outside a node are always valid. This
   is done just by keeping a hash table with the addresses, which the local
   GCs will see when it does its sweep. */

#include "backend.h"
#include "gp-type.h"

static void **dgc_hash;

#define DGC_HASH_SIZE 8209       /* Some prime number. Whatever. */

/* There's no unified global pointer type, so I need to make up a fake example
   and coerce all the incoming global pointers into that type. */
typedef char dgc_dummy;
GP_type(dgc_dummy);
extern int GC_dont_gc;

void dgc_init(void) {
  int i;
  dgc_hash = (void *)malloc(DGC_HASH_SIZE * sizeof(void *));
  for (i = 0; i < DGC_HASH_SIZE; i++) {
    dgc_hash[i] = (void *)-1;
  }
}

/* Treats addrblock as a pointer to a contiguous block of global pointers, 
   all of which must be hashed into the hash table. */
void dgc_ptr_esc(void *addrblock, int size) {
  unsigned int i,j;
  struct Pdgc_dummy *gptr_arr = (struct Pdgc_dummy *)addrblock;

  if (GC_dont_gc) return;
  /* Open addressing. */
  for (i = 0; i < size; i++) {
    void *ptr = (void *)TO_LOCAL(gptr_arr[i]);
    int proc = TO_PROC(gptr_arr[i]);
    unsigned int hashkey;
    if (proc != MYPROC) continue;
    hashkey = (unsigned int)ptr % DGC_HASH_SIZE;
    j = hashkey;
    while ((int)dgc_hash[j] != -1 && dgc_hash[j] != ptr) {
      j++;
      if (j >= DGC_HASH_SIZE) {
	j = 0;
      }
      if (j == hashkey) {
	printf("Dist. GC hash table full. Get somebody to tweak ti-dgc.c.\n");
	exit(-1);
      }
    }
    dgc_hash[j] = ptr;
  }
}

/* Used when we don't know what the global pointers are---hash everything. */
void dgc_ptr_esc_all(char *addrblock, int size) {
  void **lastptr = (void **)addrblock + (size / sizeof(void *)) - 1;
  void **ptr;
  int j;

  if (GC_dont_gc) return;
  /* Open addressing. */
  for (ptr = (void **)addrblock; ptr <= lastptr; ptr++) {
    void *val = *ptr;
    unsigned int hashkey = (unsigned int)val % DGC_HASH_SIZE;
    j = hashkey;
    while ((int)dgc_hash[j] != -1 && dgc_hash[j] != val) {
      j++;
      if (j >= DGC_HASH_SIZE) {
	j = 0;
      }
      if (j == hashkey) {
	printf("Dist. GC hash table full. Get somebody to tweak ti-dgc.c.\n");
	exit(-1);
      }
    }
    dgc_hash[j] = val;
  }
}

