/* 									tab:8
 *
 * tic.h - Support for Tic extensions
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Arvind Krishnamurthy
 * Version:	33
 * Filename:	tic.h
 * History:
 *	ssl	20	Mon Mar 21 09:46:28 1994
 *		sync_ctr for HPAM can use poll but PMAM cannot
 *	ssl	19	Sun Mar 20 14:25:01 1994
 *		sync_ctr on hps can now use poll
 *	ssl	18	Fri Mar 11 16:10:16 1994
 *		sync_ctr must use am_wait_until_zero for hpam. am_poll does not
 *      retransmit
 *	ssl	15	Thu Nov 18 18:39:21 1993
 *		Moved include of print.h to cm5.h
 *	ssl	14	Tue Nov 16 11:48:02 1993
 *		Added macros for spread and global tags to allow compiling with native
 *      compiler.
 *	ssl	13	Thu Oct 28 20:30:03 1993
 *		Removed extra prototypes. Put cm5 specific stuff in cm5.h.
 *	LTL	12	Mon Aug 16 15:52:35 1993
 *		Fixed it up for version 112 to support CMMD
 *	ACD	10	Mon Jun 14 10:56:03 1993
 *		Changed spread_free interface
 *  SL  9   Wed May 12 23:23:40 PST 1993
 *      Added SC_PROFILE option
 * 	TvE	8	Sat Apr  3 22:50:44 PST 1993
 *		Changed include <cmam/cmam.h> to "cmam.h"
 *	SL	7	Wed Feb 10 16:25:30 1993
 *		Added debugging commands (SC_DEBUG).
 *	SL	6	Wed Jan 20 12:58:42 1993
 *		Added timer library prototypes.
 * 	TvE	5	Mon Jan 18 22:55:24 PST 1993
 *		Nuked barrier.h and added #define barrier just in front of
 *		#include of cmam.h.
 *		Added declaration of all_set_PROCS. Definition is in tic.sc
 *		Changed definitions of PROCS, PROCSLOG, etc... slightly
 * 	TvE	4	Mon Jan  4 17:51:14 PST 1993
 *		Moved definitions of EXTERN and INLINE around because they
 *		conflicted with cmam.h.
 *		Added more public definitions and declarations.
 * 	TvE	3	Sat Dec 19 00:16:26 PST 1992
 *		Started major clean-up and overhaul
 *	SCG	1	Thu Dec  3 11:25:41 1992
 *		added fread and fseek
 * 	TvE	2	Tue Sep 29 12:43:37 PDT 1992
 *		include file name changes
 *	SCG	1	Wed Sep 23 00:11:02 1992
 *		made it compatible with tic.internal.h
 */

#ifndef TIC_H
#define TIC_H

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>

#if 0
#ifndef __GNUC__
#include <ansi-c.h>
#endif
#endif

/* .............................
 *
 * Control `extern' and `inline'
 *
 * .............................
 */


/* Usage: 
 * The proper (portable) way to declare inline functions in tic headers is:
 *
 *   EXTERN INLINE_MODIFIER int foo(int bar) 
 *   INLINE ({
 *     function body...
 *     })
 */

/* start with a clean slate */
#ifdef EXTERN
#undef EXTERN
#endif
#ifdef INLINE
#undef INLINE
#endif

/* Defining TIC_DEBUG has the effect of disabling all inlining of functions
 * which helps when debugging. */
#if defined(TIC_DEBUG)
#	define	INLINE(code) ;
#else
#	define	INLINE(code) {code}
#endif

/* Defining TIC_INTERNAL_DEFINE has the effect of producing a
 * real (non-extern) version of all functions and variables. 
 *
 * Exactly one .c file should define this in the distribution to 
 * support the TIC_DEBUG option and non-GNU compilers
 */
#ifdef TIC_INTERNAL_DEFINE
#      define EXTERN
#      undef INLINE
#      define INLINE(code) {code}
#else
#      define EXTERN extern
#endif

/* A hack for dealing with a native compiler that does not */
/* inline. - ssl */
#if !defined(__GNUC__) && !defined(TIC_INTERNAL_DEFINE)
#       undef   INLINE
#	define	INLINE(code) ;
#endif

#include "runtime-options-inl.h" /* INLINE_MODIFIER */


#include "backend-defines.h"
#include <primitives.h>
#include "gp-type.h"


/* ............................
 *
 * Control debugging statements
 *
 * ............................
 */

#ifdef TIC_DEBUG
#define SC_DEBUG(a) a
#else
#define SC_DEBUG(a)
#endif

/* ..........
 *
 * Boxes and Processors
 *
 * ..........
 */

#define MAX_BOX_PROCS 1024
#define MAX_PROCS 1024

#ifdef MEMORY_SHARED
#define ___MYPROC       myProcessNumber()
#define ___MYBOXPROC    myBoxProcessNumber()
extern int ___MYBOXPROCS;
#else
extern Process ___MYPROC;
#define ___MYBOXPROC    0
#define ___MYBOXPROCS  	1
#endif
extern int ___PROCS;
extern int ___BOXES;
extern Box ___MYBOX;

/* extern __tic_thread_key_t thread_key; */

#define MYPROC     ___MYPROC
#define PROCS      ___PROCS
#define MYBOX      ___MYBOX
#define MYBOXPROC  ___MYBOXPROC
#define MYBOXPROCS ___MYBOXPROCS
#define BOXES      ___BOXES

extern int ___AM_MAX_SHORT; /* size of short packet supported by AM */
extern int ___AM_MAX_MEDIUM; /* size of short packet supported by AM */
extern int ___AM_MAX_LONG; /* size of short packet supported by AM */

extern void *tic_main(int argc, char **argv);

/* ..........
 *
 * GC stuff - MUST come before pthread stuff
 *
 * ..........
 */


#ifndef USE_GC_NONE
#include "gc/gc.h"
#endif

/* ..........
 *
 * Thread operation wrappers
 *
 * ..........
 */

#define EXTRA_SYNC_CHECKING 0 /* set to non-zero to enable extra checking for error values */


#if defined (THREAD)  /* Solaris Threads */
#include <thread.h>

#define __tic_thread_t thread_t
#define __tic_thread_self() thread_self()
#define __tic_create_thread(func, arg, threadptr, result)  \
  do {   \
    if (COMM_GetMyProcsPerProc() <= 1.0) { \
      result = thr_create(NULL, NULL, func, arg, 0, threadptr); \
    } else { \
      result = thr_create(NULL, NULL, func, arg, THR_BOUND, threadptr); \
    } \
  } while(0) 
#define __tic_thread_join(thread, retval)  thr_join(0, NULL, retval)
#define __tic_thread_set_concurrency(conc) thr_setconcurrency(conc)

#define __tic_thread_key_t thread_key_t
#define __tic_thread_key_init(key)     thr_keycreate(key, NULL)
#define __tic_thread_get_key_value(key, ptr)  thr_getspecific(key, (void **)ptr)
#define __tic_thread_set_key_value(key, ptr)  thr_setspecific(key, (void *)(ptr))

#elif defined(PTHREAD)

#include <pthread.h>

#define __tic_thread_t pthread_t
#define __tic_thread_self() pthread_self()
#define __tic_create_thread(func, arg, threadptr, result)  \
  do {    \
    pthread_attr_t attr;    \
    pthread_attr_init(&attr);   \
    if (COMM_GetMyProcsPerProc() <= 1.0) { \
       pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM); \
    } \
    result = pthread_create(threadptr, &attr, func, arg);    \
  } while (0)
#define __tic_thread_join(thread, retval)  pthread_join(thread, retval)
#ifdef HAVE_PTHREAD_SETCONCURRENCY
/* this isn't supported in all pthread implementations - 
   however, it's just a hint to the pthread scheduler anyhow */
  #define __tic_thread_set_concurrency(conc) pthread_setconcurrency(conc)
#else
  #define __tic_thread_set_concurrency(conc) 
#endif

#define __tic_thread_key_t pthread_key_t 
#define __tic_thread_key_init(key)     pthread_key_create(key, NULL)
#define __tic_thread_get_key_value(key, ptr)  (*ptr) = pthread_getspecific(key)
#define __tic_thread_set_key_value(key, ptr)  pthread_setspecific(key, ptr)

#else /* NO Threads */

#define __tic_thread_t  int
#define __tic_thread_self() 0
#define __tic_create_thread(func, arg, threadptr, result) (result=0)
#define __tic_thread_join(thread, retval)  (retval?(*(void**)retval)=NULL,0:0)
#define __tic_thread_set_concurrency(conc) 0

#define __tic_thread_key_t void **
#define __tic_thread_key_init(key)  ((*key) = (void **)malloc(sizeof(void *)), (int)!(*key))
#define __tic_thread_get_key_value(key, ptr)  ((*(void **)ptr) = (*key), 0)
#define __tic_thread_set_key_value(key, ptr)  ((*key) = (void *)ptr, 0)


#endif /* Threads */


/* Locks and Mutexes */

#if defined(IRIX_THREADS) 
#define CAUTIOUS_MONITORS
#endif

#ifndef MEMORY_SHARED

#define __tic_lock_t int
#define __tic_cond_t int

#define __tic_lock_init(lock)      
#define __tic_lock_initializer  0
#define __tic_lock_decl_initializer 0
#define __tic_cond_init(cond, lock)
#define __tic_cond_initializer  0
#define __tic_cond_decl_initializer 0

#define __tic_lock(lock)           
#define __tic_try_lock(lock) 1          
#define __tic_unlock(lock)         
#define __tic_mon_wait(mon)           abort()
#define __tic_mon_timedwait(mon, tm)  abort()
#define __tic_mon_signal(mon)         abort()
#define __tic_mon_broadcast(mon)      abort()

#define __tic_lock_destroy(lock)      
#define __tic_cond_destroy(cond)

#define ___tic_time_t	struct timespec

#else

/* error checking macros for thread functions */
#if EXTRA_SYNC_CHECKING
  EXTERN INLINE_MODIFIER int __tic_errcheck_helper(int actualval, int okval, int altok, char* where) 
  INLINE({                                  
    if ((!!actualval) == okval || actualval == altok) return actualval;
    else {
      char altstr[20];
      if (altok != okval) sprintf(altstr, "(or %i)", altok);
      else altstr[0] = '\0';
      fprintf(stderr, "Tic: Error while calling: %s\n", where);  
      fprintf(stderr, "Tic: expected %s%s, got %i\n", 
        (okval?"non-zero":"zero"), altstr, actualval); 
      perror("err code: ");
      abort();
      return actualval;
    }
  })
  #define __tic_errcheckZ(fn) __tic_errcheck_helper(fn, 0, 0, #fn)
  #define __tic_errcheckZopt(fn,opt) __tic_errcheck_helper(fn, 0, opt, #fn)
  #define __tic_errcheckNZ(fn) __tic_errcheck_helper(fn, 1, 1, #fn)
#else /* compile away */
  #define __tic_errcheckZ(fn)  (fn)
  #define __tic_errcheckZopt(fn,opt)  (fn)
  #define __tic_errcheckNZ(fn) (fn)
#endif

#if defined (THREAD)  /* Solaris Threads */

#define __tic_lock_t mutex_t
#define __tic_cond_t cond_t

/* If at runtime we know there's only one processor on this box, we 
   don't need to perform synchronization. */
#define __tic_lock_init(lock)      \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_init(lock, USYNC_THREAD, NULL)) : 0)

EXTERN INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({                                  
  __tic_lock_t var = DEFAULT_MUTEX;
  return var;                                    
})

#define __tic_lock_initializer  __tic_new_lock()
#define __tic_lock_decl_initializer DEFAULT_MUTEX

#define __tic_cond_init(cond, lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(cond_init(cond, USYNC_THREAD, NULL)) : 0)

EXTERN INLINE_MODIFIER __tic_cond_t __tic_new_cond(void) 
INLINE({                                  
  __tic_cond_t var = DEFAULTCV;
  return var;                                   
})

#define __tic_cond_initializer __tic_new_cond()
#define __tic_cond_decl_initializer DEFAULTCV


#define __tic_lock(lock)           \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_lock(lock)) : 0)
#define __tic_try_lock(lock)       \
  ((MYBOXPROCS > 1) ? __tic_errcheckZopt(mutex_trylock(lock), EBUSY) : 1)
#define __tic_unlock(lock)         \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_unlock(lock)) : 0)
#define __tic_mon_wait(mon)            \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_wait(&(mon->condvar), &(mon->lock))); } while(0)
#define __tic_mon_timedwait(mon, tm)   do {					\
	if (MYBOXPROCS > 1) \
        __tic_errcheckZopt(cond_timedwait(&(mon->condvar), &(mon->lock), tm), ETIMEDOUT);	\
	else ___tic_sleepuntil(tm);						\
	} while(0)
#define __tic_mon_signal(mon)          \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_signal(&(mon->condvar))); } while(0)
#define __tic_mon_broadcast(mon)       \
  do { if (MYBOXPROCS > 1) __tic_errcheckZ(cond_broadcast(&(mon->condvar))); } while(0)

#define __tic_lock_destroy(lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(mutex_destroy(lock)) : 0)
#define __tic_cond_destroy(cond)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(cond_destroy(cond)) : 0)

#define ___tic_time_t	timestruc_t

#elif defined(PTHREAD)

#define __tic_lock_t pthread_mutex_t
#define __tic_cond_t pthread_cond_t

/* If at runtime we know there's only one processor on this box, we 
   don't need to perform synchronization. */
#if EXTRA_SYNC_CHECKING && defined(LINUX_THREADS)
EXTERN INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({      
  __tic_lock_t lock;
  pthread_mutexattr_t attr; 
  __tic_errcheckZ(pthread_mutexattr_init(&attr));
  __tic_errcheckZ(pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_ERRORCHECK_NP));
  __tic_errcheckZ(pthread_mutex_init(&lock, &attr));
  __tic_errcheckZ(pthread_mutexattr_destroy(&attr));
  return lock;
  })
#define __tic_lock_init(lock)   \
   ((MYBOXPROCS > 1) ? *lock = __tic_new_lock(), 0 : 0)
#else
#define __tic_lock_init(lock)      \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_init(lock, NULL)) : 0)
EXTERN INLINE_MODIFIER __tic_lock_t __tic_new_lock(void) 
INLINE({      
  __tic_lock_t var = PTHREAD_MUTEX_INITIALIZER;  
  return var;                                    
})
#endif

#define __tic_lock_decl_initializer  PTHREAD_MUTEX_INITIALIZER
#define __tic_lock_initializer  __tic_new_lock()

#define __tic_cond_init(cond, lock)   \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_cond_init(cond, NULL)) : 0)

EXTERN INLINE_MODIFIER __tic_cond_t __tic_new_cond(void) 
INLINE({                                  
  __tic_cond_t var = PTHREAD_COND_INITIALIZER;  
  return var;                                   
})

#define __tic_cond_initializer __tic_new_cond()
#define __tic_cond_decl_initializer PTHREAD_COND_INITIALIZER

#define __tic_lock(lock)           \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_lock(lock)) : 0)
#define __tic_try_lock(lock)       \
  ((MYBOXPROCS > 1) ? __tic_errcheckZopt(pthread_mutex_trylock(lock), EBUSY) : 1)
#define __tic_unlock(lock)         \
  ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_unlock(lock)) : 0)

#ifdef CAUTIOUS_MONITORS
/* some pthread implementations need special help for wait/notify -
   specifically, IRIX pthreads frequently return prematurely from a wait
      so we must explicitly check the condition upon return 
   the extra fields in the monitor data structure are protected by the lock,
   which already must be held in order to call these wait/notify functions */
#define __tic_mon_wait(mon)   do { if (MYBOXPROCS > 1) {   		\
  mon->numwaiting++; /* about to wait */				\
  do {									\
   __tic_errcheckZ(pthread_cond_wait(&(mon->condvar), &(mon->lock)));   \
   } while (!mon->numsignaled);						\
  mon->numwaiting--; /* we're no longer waiting */			\
  mon->numsignaled--; /* we're awaking - consume one signal */ 		\
  } } while(0)
#define __tic_mon_timedwait(mon, tm)   do { if (MYBOXPROCS > 1) {   	\
  int retval = 0;							\
  mon->numwaiting++; /* about to wait */				\
  do {									\
   retval = __tic_errcheckZopt(pthread_cond_timedwait(&(mon->condvar), &(mon->lock), tm), ETIMEDOUT);	\
   } while (!mon->numsignaled && retval != ETIMEDOUT);			\
  mon->numwaiting--; /* we're no longer waiting */			\
  if (retval != ETIMEDOUT) mon->numsignaled--; /* we're awaking - consume one signal */ \
  } else ___tic_sleepuntil(tm); } while(0)
#define __tic_mon_signal(mon)   do { if (MYBOXPROCS > 1) {  		\
  if (mon->numwaiting > mon->numsignaled) /* some waiter still needs a signal */ \
    mon->numsignaled++; /* add one signal */				\
  __tic_errcheckZ(pthread_cond_signal(&(mon->condvar)));		\
  } } while(0)
#define __tic_mon_broadcast(mon)   do { if (MYBOXPROCS > 1) {  		\
  mon->numsignaled = mon->numwaiting; /* ensure a signal for each current waiter */ \
  __tic_errcheckZ(pthread_cond_broadcast(&(mon->condvar)));		\
  } } while(0)
#else
#define __tic_mon_wait(mon)   \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_wait(&(mon->condvar), &(mon->lock))); } while (0)
#define __tic_mon_timedwait(mon, tm)   do {						\
    if (MYBOXPROCS > 1) __tic_errcheckZopt(pthread_cond_timedwait(&(mon->condvar), &(mon->lock), tm),ETIMEDOUT);	\
    else  ___tic_sleepuntil(tm);							\
    } while (0)
#define __tic_mon_signal(mon)    \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_signal(&(mon->condvar))); } while (0)
#define __tic_mon_broadcast(mon) \
   do { if (MYBOXPROCS > 1) __tic_errcheckZ(pthread_cond_broadcast(&(mon->condvar))); } while (0)
#endif

#define __tic_lock_destroy(lock)   \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_mutex_destroy(lock)) : 0)
#define __tic_cond_destroy(cond)   \
   ((MYBOXPROCS > 1) ? __tic_errcheckZ(pthread_cond_destroy(cond)) : 0)

#define ___tic_time_t	struct timespec

#endif

#endif /* NO_SMP For machines like the NOW */  

/* Timer-related macros/functions */
/* ___tic_time_t is defined separately for each platform above */

/* set a time structure for given millis from right now */
#define ___tic_set_wait_timer(tm, millis) do { 				\
    struct timeval now;							\
    gettimeofday(&now, NULL);						\
    (tm)->tv_sec = now.tv_sec;						\
    (tm)->tv_sec += (now.tv_usec + millis*1000)/1000000;		\
    (tm)->tv_nsec = ((now.tv_usec + millis*1000)%1000000) * 1000;	\
    } while (0)

/* block until the given absolute time has approximately been reached */
#define ___tic_sleepuntil(tm) do {				\
	int millis;						\
	struct timeval now;					\
	gettimeofday(&now, NULL);				\
	millis = ((tm)->tv_sec - now.tv_sec) * 1000 +		\
	         ((tm)->tv_nsec / 1000 - now.tv_usec) / 1000;	\
	if (millis > 0) m5sleepJmT6Thread4lang4java(millis);	\
	} while (0)

/* return the number of milliseconds from now until the time indicated by tm
 * (will be negative if the time indicated by tm has already passed)
 */
EXTERN INLINE_MODIFIER int ___tic_millisuntil(___tic_time_t* tm)
INLINE ({
    int millis;                                             
    struct timeval now;                                     
    gettimeofday(&now, NULL);                               
    millis = ((tm)->tv_sec - now.tv_sec) * 1000 +           
             ((tm)->tv_nsec / 1000 - now.tv_usec) / 1000;        
    return millis;
  })

#if defined(GLUNIX)
#include <glib.h>
#endif

#include <unistd.h>
#include <comm_utils.h>

extern __tic_thread_key_t tic_thread_key;

processInfo *myProcess() __attribute__ ((const));
EXTERN INLINE_MODIFIER processInfo *myProcess()
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi);
})


Process myProcessNumber() __attribute__ ((const));
EXTERN INLINE_MODIFIER Process myProcessNumber() 
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi)->processNumber;
})

Process myBoxProcessNumber() __attribute__ ((const));
EXTERN INLINE_MODIFIER Process myBoxProcessNumber() 
INLINE({
  processInfo *pi; 
  __tic_thread_get_key_value(tic_thread_key, &pi);
  return (pi)->boxProcessNumber;
})

EXTERN __tic_lock_t AM_lock;

/* ........................
 *
 * Message Layer
 *
 * ........................
 */

#define LOWWORD(i)    ((julong)(i) & 0xFFFFFFFFUL)
#define HIGHWORD(i)   LOWWORD((julong)(i) >> 32) 

#define SETLOWWORD(ptr, i)  (*(ptr)) = (HIGHWORD(*(ptr)) << 32) | LOWWORD(i)
#define SETHIGHWORD(ptr, i) (*(ptr)) = (LOWWORD(((julong)(i))) << 32) | LOWWORD(*(ptr))


#ifdef COMM_AM2
#if defined(COMM_AMUDP)
#include <amudp.h>
#include <amudp_spmd.h>
#elif defined(COMM_AMLAPI)
#include <amlapi.h>
#elif defined(SOLARIS)
#include <am.h>
#elif defined(LINUX)
#include <amvia.h>
#endif

#include <tic_am2_macros.h>
 
#define TIC_TAG		18

extern eb_t __sc_bundle;
extern ep_t __sc_endpoint;
#define ADVANCE_SEND_ENDPOINT
#define CUR_ENDPOINT          __sc_endpoint

#endif /* AMII */

/* ........................
 *
 * Synchronization counters
 *
 * ........................
 */
typedef volatile int Counter;


#ifdef COMM_AM2
EXTERN INLINE_MODIFIER void tic_poll()
INLINE({
  __tic_lock(&AM_lock);
	AM_Poll(__sc_bundle);
  __tic_unlock(&AM_lock);
})
#else /* AMII */
#define tic_poll()
#endif

#ifdef COMM_AM2
EXTERN INLINE_MODIFIER void sync_ctr(Counter *ctr)
INLINE({
	while(*ctr) {
	  tic_poll();
	}
})
#else
#define sync_ctr(a) 
#endif /* AMII */

#ifdef COMM_AM2
EXTERN INLINE_MODIFIER int is_sync_ctr(Counter *ctr)
INLINE({
  tic_poll(); 
  return !*ctr;
})
#else
#define is_sync_ctr(a)  (1)
#endif /* AMII */


extern int politep;


/* ...............
 *
 * Global pointers
 *
 * ...............
 */

#include <pointers.h>


/* .............
 *
 * Handlers
 *
 * .............
 */

#include <handlers.h>

/* .............
 *
 * Global memory
 *
 * .............
 */

/* Constants specifying when to use segments
 * (these should go somewhere else... -TvE */
#define		STORE_STRUCT_LEN	60
#define		PUT_STRUCT_LEN		32

/* Ack-ed reads and writes */
#include <memory.h>

/* Non-ack-ed stores */
#include <store.h>

#include <bulk.h>

/*************************************************************
 *
 * Global communication
 * 
 *************************************************************/
#include <com.h>

/* getenvMaster() retrieves environment information from the console node */
#ifdef COMM_AMUDP
  extern int usingAMUDPSPMD;
  EXTERN INLINE_MODIFIER const char *getenvMaster(const char *name)
  INLINE({
    if (usingAMUDPSPMD) return AMUDP_SPMDgetenvMaster(name);
    else return getenv(name);
  })
#else
  EXTERN INLINE_MODIFIER const char *getenvMaster(const char *name)
  INLINE({
    return getenv(name);
  })
#endif

/* ..............
 *
 * Global control
 *
 * ..............
 */

#include <control.h> /* public stuff. Maybe should be here verbatim? */

/* ........................
 *
 * Barrier
 *
 * ........................
 */
#include <barrier.h>

/* ........................
 *
 * Monitors
 *
 * ........................
 */

#include <monitor.h>


/* .........................
 *
 * Timer library
 *
 * .........................
 */

EXTERN double get_seconds();


#endif
