/* 									tab:8
 *
 * tic.h - Support for Tic extensions
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Arvind Krishnamurthy
 * Version:	33
 * Filename:	tic.h
 * History:
 *	ssl	20	Mon Mar 21 09:46:28 1994
 *		sync_ctr for HPAM can use poll but PMAM cannot
 *	ssl	19	Sun Mar 20 14:25:01 1994
 *		sync_ctr on hps can now use poll
 *	ssl	18	Fri Mar 11 16:10:16 1994
 *		sync_ctr must use am_wait_until_zero for hpam. am_poll does not
 *      retransmit
 *	ssl	15	Thu Nov 18 18:39:21 1993
 *		Moved include of print.h to cm5.h
 *	ssl	14	Tue Nov 16 11:48:02 1993
 *		Added macros for spread and global tags to allow compiling with native
 *      compiler.
 *	ssl	13	Thu Oct 28 20:30:03 1993
 *		Removed extra prototypes. Put cm5 specific stuff in cm5.h.
 *	LTL	12	Mon Aug 16 15:52:35 1993
 *		Fixed it up for version 112 to support CMMD
 *	ACD	10	Mon Jun 14 10:56:03 1993
 *		Changed spread_free interface
 *  SL  9   Wed May 12 23:23:40 PST 1993
 *      Added SC_PROFILE option
 * 	TvE	8	Sat Apr  3 22:50:44 PST 1993
 *		Changed include <cmam/cmam.h> to "cmam.h"
 *	SL	7	Wed Feb 10 16:25:30 1993
 *		Added debugging commands (SC_DEBUG).
 *	SL	6	Wed Jan 20 12:58:42 1993
 *		Added timer library prototypes.
 * 	TvE	5	Mon Jan 18 22:55:24 PST 1993
 *		Nuked barrier.h and added #define barrier just in front of
 *		#include of cmam.h.
 *		Added declaration of all_set_PROCS. Definition is in tic.sc
 *		Changed definitions of PROCS, PROCSLOG, etc... slightly
 * 	TvE	4	Mon Jan  4 17:51:14 PST 1993
 *		Moved definitions of EXTERN and INLINE around because they
 *		conflicted with cmam.h.
 *		Added more public definitions and declarations.
 * 	TvE	3	Sat Dec 19 00:16:26 PST 1992
 *		Started major clean-up and overhaul
 *	SCG	1	Thu Dec  3 11:25:41 1992
 *		added fread and fseek
 * 	TvE	2	Tue Sep 29 12:43:37 PDT 1992
 *		include file name changes
 *	SCG	1	Wed Sep 23 00:11:02 1992
 *		made it compatible with tic.internal.h
 */

#ifndef _TIC_AM_H
#define _TIC_AM_H

#include <primitives.h>

/* ------------------------------------------------------------------------------------ */

#ifndef COMM_AMUDP

#define AM_Request1(endpoint, dest, handler, a) \
      AM_Request4(endpoint, dest, handler, (int)a, 0, 0, 0)

#define AM_Request2(endpoint, dest, handler, a, b) \
      AM_Request4(endpoint, dest, handler, (int)a, (int)b, 0, 0)

#define AM_Request3(endpoint, dest, handler, a, b, c) \
      AM_Request4(endpoint, dest, handler, (int)a, (int)b, (int)c, 0)

#define AM_Request5(endpoint, dest, handler, a, b, c, d, e) \
      AM_Request8(endpoint, dest, handler, (int)a, (int)b, (int)c, (int)d, \
				  (int)e, 0, 0, 0)

#define AM_Request6(endpoint, dest, handler, a, b, c, d, e, f) \
      AM_Request8(endpoint, dest, handler, (int)a, (int)b, (int)c, (int)d, \
				  (int)e, (int)f, 0, 0)

#define AM_Request7(endpoint, dest, handler, a, b, c, d, e, f, g) \
      AM_Request8(endpoint, dest, handler, (int)a, (int)b, (int)c, (int)d, \
				  (int)e, (int)f, (int)g, 0)

#define AM_Reply1(token, handler, a) \
      AM_Reply4(token, handler, (int)a, 0, 0, 0)

#define AM_Reply2(token, handler, a, b) \
      AM_Reply4(token, handler, (int)a, (int)b, 0, 0)

#define AM_Reply3(token, handler, a, b, c) \
      AM_Reply4(token, handler, (int)a, (int)b, (int)c, 0)

#endif

#ifdef COMM_VIA

#define AM_RequestXfer4(request_endpoint, reply_endpoint, dest_offset,   \
                        handler, source_addr, nbytes, arg0, arg1,        \
                        arg2, arg3)                                      \
        AM_RequestXfer8(request_endpoint, reply_endpoint, dest_offset,   \
                        handler, source_addr, nbytes, arg0, arg1, arg2,  \
                        arg3, 0, 0, 0, 0)

#define AM_ReplyXfer4(token, dest_offset, handler, source_addr,  \
                      nbytes, arg0, arg1, arg2, arg3)            \
        AM_ReplyXfer8(token, dest_offset, handler, source_addr,  \
                      nbytes, arg0, arg1, arg2, arg3, 0, 0, 0, 0)
#endif /* COMM_VIA */


#define AM_Safe(__function) 				\
do { 	int __rc; 								\
	if ((__rc = __function) != AM_OK) { 	\
		fprintf(stderr, "ERROR: %s failed with code %d\n", #__function, __rc); \
	    abort(); 							\
	} 										\
  } while(0)
									  
/* ------------------------------------------------------------------------------------ */
#ifdef PTR64


#define TI_AM_Request1(a,b,c,d) \
  do {   \
  __tic_lock(&AM_lock);         \
     AM_Safe(AM_Request2(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d)));  \
  __tic_unlock(&AM_lock);       \
       } while (0)
#define TI_AM_Request1(a,b,c,d,e) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_Request4(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d), \
                          (int)HIGHWORD(e), (int)LOWWORD(e))); \
  __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request3(a,b,c,d,e,f) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_Request6(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f))); \
  __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request4(a,b,c,d,e,f,g) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_Request8(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f), \
                 (int)HIGHWORD(g), (int)LOWWORD(g))); \
  __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestI2(a,b,c,d,e,f,g) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestI4(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f), \
                 (int)HIGHWORD(g), (int)LOWWORD(g))); \
  __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestI4(a,b,c,d,e,f,g,h,i) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestI8(a, b, c, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f), \
                 (int)HIGHWORD(g), (int)LOWWORD(g), \
                 (int)HIGHWORD(h), (int)LOWWORD(h), \
                 (int)HIGHWORD(i), (int)LOWWORD(i))); \
  __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Reply0(a, b)   \
     AM_Safe(AM_Reply0(a, b))

#define TI_AM_Reply1(a, b, d) \
     AM_Safe(AM_Reply2(a, b, (int)HIGHWORD(d), (int)LOWWORD(d)))

#define TI_AM_Reply2(a, b, d, e) \
     AM_Safe(AM_Reply4(a, b, (int)HIGHWORD(d), (int)LOWWORD(d), \
                     (int)HIGHWORD(e), (int)LOWWORD(e))

#define TI_AM_Reply3(a, b, d, e, f) \
     AM_Safe(AM_Reply6(a, b, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f)))

#define TI_AM_Reply4(a, b, d, e, f, g) \
     AM_Safe(AM_Reply8(a, b, (int)HIGHWORD(d), (int)LOWWORD(d), \
                 (int)HIGHWORD(e), (int)LOWWORD(e), \
                 (int)HIGHWORD(f), (int)LOWWORD(f), \
                 (int)HIGHWORD(g), (int)LOWWORD(g)))

#define TI_AM_ReplyI2(a, b, d, e, f, g) \
     AM_Safe(AM_ReplyI4(a, b, (int)HIGHWORD(d), (int)LOWWORD(d), \
                     (int)HIGHWORD(e), (int)LOWWORD(e), \
                     (int)HIGHWORD(f), (int)LOWWORD(f), \
                     (int)HIGHWORD(g), (int)LOWWORD(g)))

#define TI_AM_ReplyI4(a, b, d, e, f, g) \
     AM_Safe(AM_ReplyI8(a, b, (int)HIGHWORD(d), (int)LOWWORD(d), \
                     (int)HIGHWORD(e), (int)LOWWORD(e), \
                     (int)HIGHWORD(f), (int)LOWWORD(f), \
                     (int)HIGHWORD(g), (int)LOWWORD(g), \
                     (int)HIGHWORD(h), (int)LOWWORD(h), \
                     (int)HIGHWORD(i), (int)LOWWORD(i)))

#define TI_AM_RequestXfer0(a, b, c, d, e, f) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestXfer4(a, b, (int)c, d, e, (int)f, 0, 0, 0, 0)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestXfer1(a, b, c, d, e, f, g) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestXfer4(a, b, (int)c, d, e, f, \
                     (int)HIGHWORD(g), (int)LOWWORD(g), 0, 0)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_ReplyXfer1(token, dest_offset, handler, src_addr, len, f) \
     AM_Safe(AM_ReplyXfer4(token, (int)dest_offset, handler, (void *)src_addr, len, \
                     (int)HIGHWORD(f), (int)LOWWORD(f), 0, 0))


/* ------------------------------------------------------------------------------------ */
/* ^^^ PTR64  ,,, PTR32 */
#else

#define TI_AM_Request1(a, b, c, d) \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_Request1(a, b, c, d)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request2(a, b, c, d, e)  \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_Request2(a, b, c, d, e)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request3(a, b, c, d, e, f) \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_Request3(a, b, c, d, e, f)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request4(a, b, c, d, e, f, g)  \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_Request4(a, b, c, d, e, f, g)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Request5(a, b, c, d, e, f, g, h) \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_Request5(a, b, c, d, e, f, g, h)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestI2(a, b, c, d, e, f, g)  \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_RequestI4(a, b, c, d, e, (int)f, (int)g, 0, 0)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestI4(a, b, c, d, e, f, g, h, i)  \
  do {                            \
  __tic_lock(&AM_lock);           \
  AM_Safe(AM_RequestI4(a, b, c, d, e, (int)f, (int)g, (int)h, (int)i)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_Reply0(a, b)    AM_Safe(AM_Reply0(a, b))
#define TI_AM_Reply1(a, b, d) AM_Safe(AM_Reply1(a, b, d))
#define TI_AM_Reply2(a, b, d, e) AM_Safe(AM_Reply2(a, b, d, e))
#define TI_AM_Reply3(a, b, d, e, f) AM_Safe(AM_Reply3(a, b, d, e, f))
#define TI_AM_Reply4(a, b, d, e, f, g) AM_Safe(AM_Reply4(a, b, d, e, f, g))

#define TI_AM_ReplyI2(a, b, d, e, f, g) AM_Safe(AM_ReplyI4(a, b, d, e, (int)f, (int)g, 0, 0))
#define TI_AM_ReplyI4(a, b, d, e, f, g, h, i) AM_Safe(AM_ReplyI4(a, b, d, e, (int)f, (int)g, (int)h, (int)i))

#define TI_AM_RequestXfer0(a, b, c, d, e, f) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestXfer4(a, b, c, d, e, f, 0, 0, 0, 0)); \
   __tic_unlock(&AM_lock);         \
       } while (0)

#define TI_AM_RequestXfer1(a, b, c, d, e, f, g) \
  do {                            \
  __tic_lock(&AM_lock);           \
     AM_Safe(AM_RequestXfer4(a, b, c, d, e, f, g, 0, 0, 0)); \
   __tic_unlock(&AM_lock);         \
       } while (0)


#define TI_AM_ReplyXfer1(a, b, c, d, e, f) \
     AM_Safe(AM_ReplyXfer4(a, b, c, (void *)d, e, f, 0, 0, 0))

#endif /* PTR64 */





#endif /* _TIC_AM_H */

