/* 									tab:8
 *
 * memory.sc - Global memory accesses with acks (get, put, read and write)
 * 
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author:	Thorsten von Eicken
 * Version:	7
 * Filename:	memory.sc
 * History:
 *	ssl	5	Sun Dec  5 22:59:24 1993
 *		am and handlers now use unsigned instead of long long as arguments.
 *      doubles will be passed in fp registers and long longs will be dword
 *      aligned.
 *	ssl	4	Tue Nov 16 12:21:40 1993
 *	  Replaced global tags with macros.
 * 	TvE	3	Wed Dec 23 14:41:44 PST 1992
 *		Create from tic.sc.
 */

#include <tic.h>

__tic_lock_t	GetPut_mutex;

void memory_init(void)
{
  if (MYBOXPROC == 0) {
    __tic_lock_init(&GetPut_mutex);
  }
}

/* .......................................
 *
 * Keep track of outstanding GETs and PUTs
 *
 * .......................................
 */

Counter numGetOuts = 0;
Counter numPutOuts = 0;

void __decr_put_ctr(void) { 
  __tic_lock(&GetPut_mutex);
  numPutOuts--; 
  __tic_unlock(&GetPut_mutex);
}

void __decr_put_ctr_reply_handler(void *token) { 
  __tic_lock(&GetPut_mutex);
  numPutOuts--; 
  __tic_unlock(&GetPut_mutex);
}
	

/* ..........................
 *
 * GETs with explicit counter
 *
 * ..........................
 */

#ifdef COMM_AM2


___GET_HANDLER(b, jbyte)
___GET_HANDLER(sh, jshort)
___GET_HANDLER(i, jint)
___GET_HANDLER(f, jfloat)
___GET_HANDLER(d, jdouble)
___GET_HANDLER(l, jlong)
___GET_HANDLER(lp, void *)
___GET_HANDLER(gp, jGPointer)


void __b_get_handler(void *token, jbyte *addr, jbyte *retAddr, Counter *ctr)
{
  TI_AM_Reply3(token, __B_GET_RH, (jUIntPointer)retAddr, 
	       *addr, (jUIntPointer)ctr);
}

void __sh_get_handler(void *token, jshort *addr, jshort *retAddr, Counter *ctr)
{
  TI_AM_Reply3(token, __SH_GET_RH, (jUIntPointer)retAddr, 
	       *addr, (jUIntPointer)ctr);
}

void __i_get_handler(void *token, jint *addr, jint *retAddr, Counter *ctr)
{
  TI_AM_Reply3(token, __I_GET_RH, (jUIntPointer)retAddr, 
	       *addr, (jUIntPointer)ctr);
}

void __f_get_handler(void *token, jfloat *addr, jfloat *retAddr, Counter *ctr)
{
  TI_AM_Reply3(token, __F_GET_RH, (jUIntPointer)retAddr, 
	       *addr, (jUIntPointer)ctr);
}

void __d_get_handler(void *token, jdouble *addr, jdouble *retAddr, 
		     Counter *ctr)
#ifdef PTR64 
{
  /* some machines pass doubles in fp registers. force passing in int
     registers. - ssl */
  TI_AM_Reply3(token, __D_GET_RH, (jUIntPointer)retAddr, 
	       *addr, (jUIntPointer)ctr);
}
#else
{
  /* some machines pass doubles in fp registers. force passing in int
     registers. - ssl */
  TI_AM_Reply4(token, __D_GET_RH, (jUIntPointer)retAddr, 
	    HIGHWORD(*((julong *)addr)), 
	    LOWWORD(*((julong *)addr)), (jUIntPointer)ctr);
}
#endif

void __l_get_handler(void *token, jlong *addr, jlong *retAddr, Counter *ctr)
#ifdef PTR64
{
  /* pass as two unsigned so that HPs won't dword align long long - ssl */
	TI_AM_Reply3(token, __L_GET_RH, (jUIntPointer)retAddr, 
		     *addr, (jUIntPointer)ctr);
}
#else
{
  /* pass as two unsigned so that HPs won't dword align long long - ssl */
  TI_AM_Reply4(token, __L_GET_RH, (jUIntPointer)retAddr, 
	    HIGHWORD((julong)*addr), 
	    LOWWORD((julong)*addr), (jUIntPointer)ctr);
}
#endif

void __lp_get_handler(void *token, void **addr, void **retAddr, Counter *ctr)
{
  TI_AM_Reply3(token, __LP_GET_RH, (jUIntPointer)retAddr, 
	       (jUIntPointer)*addr, (jUIntPointer)ctr);
}

#ifdef PTR64
void __gp_get_handler(void *token, jGPointer *addr, jGPointer *retAddr, 
		      Counter *ctr)
{
  jUIntPointer data0, data1;

  data0 = toboxproc(*addr);
  data1 = (jUIntPointer)tolocal(*addr);
  TI_AM_Reply4(token, __GP_GET_RH, (jUIntPointer)retAddr, 
	       data0, data1, (jUIntPointer)ctr);
}
#else
void __gp_get_handler(void *token, jGPointer *addr, jGPointer *retAddr, 
		      Counter *ctr)
{
  jGPointer data = *addr;
  jUIntPointer data0  = HIGHWORD(*(julong *)&data);
  jUIntPointer data1  = LOWWORD(*(julong *)&data);
  TI_AM_Reply4(token, __GP_GET_RH, (jUIntPointer)retAddr, 
	       data0, data1, (jUIntPointer)ctr);
}
#endif


___GET_REPLY_HANDLER4(b, jbyte)
___GET_REPLY_HANDLER4(sh, jshort)
___GET_REPLY_HANDLER4(i, jint)
___GET_REPLY_HANDLER4(f, jfloat)
___GET_REPLY_HANDLER8(d, jdouble)
___GET_REPLY_HANDLER8(l, jlong)
#ifdef PTR64
     ___GET_REPLY_HANDLER16(gp, jGPointer, jUIntPointer)
     ___GET_REPLY_HANDLER8(lp, void *)
#else
     ___GET_REPLY_HANDLER8(gp, jGPointer)
     ___GET_REPLY_HANDLER4(lp, void *)
#endif


void __b_get_reply_handler(void *token, jbyte *addr, jbyte data, Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}

void __sh_get_reply_handler(void *token, jshort *addr, jshort data, Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}

void __i_get_reply_handler(void *token, jint *addr, jint data, Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}

void __f_get_reply_handler(void *token, jfloat *addr, jfloat data, Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}

/* must talk unsigned instead of double as args because some machines
   pass doubles in fp registers. - ssl */
void __d_get_reply_handler(void *token, jdouble *addr, jdouble data,
			   Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}


void __l_get_reply_handler(void *token, jlong *addr, jlong data,
			   Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}


void __lp_get_reply_handler(void *token, void **addr, void *data,
			   Counter *ctr)
{
  addr[0] = data;
  __decr_ctr(ctr);
}

#ifdef PTR64
void __gp_get_reply_handler(void *token, jGPointer *addr, 
			    jUIntPointer data0, jUIntPointer data1,
			    Counter *ctr)
{
  *addr = mutateglobalwithbox(*addr, boxproctoproc(data0), 
			      boxproctobox(data0), (void *)data1);
  __decr_ctr(ctr);
}
#else
void __gp_get_reply_handler(void *token, jGPointer *addr, 
			    jGPointer data, 
			    Counter *ctr)
{
  *addr = data;
  __decr_ctr(ctr);
}

#endif



/* ..........................
 *
 * PUTs with explicit counter
 *
 * ..........................
 */



___PUT_HANDLER4(b, jbyte)
___PUT_HANDLER4(sh, jshort)
___PUT_HANDLER4(i, jint)
___PUT_HANDLER4(f, jfloat)
___PUT_HANDLER8(d, jdouble)
___PUT_HANDLER8(l, jlong)
#ifdef PTR64
     ___PUT_HANDLER16(gp, jGPointer, jUIntPointer)
     ___PUT_HANDLER8(lp, void *)
#else
void ___gp_put_handler(void *token, juint addr, juint box, juint proc, juint addrData,
			    juint ctr) {
  jGPointer data;
  mutateglobalwithbox(data, proc, box, addrData);
  __gp_put_handler(token, (jGPointer *)addr, data, (Counter *)ctr);
}
     ___PUT_HANDLER4(lp, void *)
#endif



void __b_put_handler(void *token, jbyte *addr, jbyte data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __sh_put_handler(void *token, jshort *addr, jshort data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __i_put_handler(void *token, jint *addr, jint data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __f_put_handler(void *token, jfloat *addr, jfloat data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __d_put_handler(void *token, jdouble *addr, jdouble data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __l_put_handler(void *token, jlong *addr, jlong data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}

void __lp_put_handler(void *token, void **addr, void *data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}


#ifdef PTR64
void __gp_put_handler(void *token, jGPointer *addr, jUIntPointer data0,
		      jUIntPointer data1, Counter *ctr)
{
  *addr = mutateglobalwithbox(*addr, boxproctoproc(data0), 
			      boxproctobox(data0), (void *)data1);
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}
#else
void __gp_put_handler(void *token, jGPointer *addr,
		      jGPointer data, Counter *ctr)
{
  *addr = data;
  TI_AM_Reply1(token, __DECR_CTR_RH, (jUIntPointer)ctr);
}
#endif

#endif /* AMII */


/* ..........................
 *
 * PUTs with implicit counter
 *
 * ..........................
 */

#if 0
/* must pass as two unsigneds because some machines pass doubles in fp
   registers. - ssl */
void __d_put_implicit_handler(void *token,  double *addr, unsigned word0,
							  unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	TI_AM_Reply0(token, __DECR_PUT_CTR_RH);
	/*am_reply_0(retNode, __decr_put_ctr);*/
}

#ifdef __GNUC__
void __ll_put_implicit_handler(void *token,  long long *addr, unsigned word0,
							   unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	TI_AM_Reply0(token, __DECR_PUT_CTR_RH);
	/*am_reply_0(retNode, __decr_put_ctr);*/
}
#endif
#endif


/* ......
 *
 * WRITEs
 *
 * ......
 */

/* Special counter for double/long-long writes */
#if 0

Counter __write_ctr=0;

void __decr_write_ctr_reply_handler(void *token) {
	__tic_lock(&GetPut_mutex);
	__write_ctr = 0;
	__tic_unlock(&GetPut_mutex);
}
#endif


#if 0
/* pass as unsigned so it will not pass in fp registers */
void __d_write_handler(void *token,  double *addr, unsigned word0,
					   unsigned word1)
{
	*((unsigned*)addr) = word0;
	*((unsigned*)addr+1) = word1;
	TI_AM_Reply0(token, __DECR_WRITE_CTR_RH);
	/*am_reply_0(retNode, __decr_write_ctr);*/
}

#ifdef __GNUC__
void __ll_write_handler(void *token,  long long *addr, unsigned word0,
						unsigned word1)
{
	*((unsigned*) addr) = word0;
	*((unsigned*) addr+1) = word1;
	TI_AM_Reply0(token, __DECR_WRITE_CTR_RH);
	/*am_reply_0(retNode, __decr_write_ctr);*/
}
#endif
#endif






