/*									tab:8
 *
 * pointers.h - Make global pointers happen. Compiler specific stuff.
 *
 * "Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 		Thorsten von Eicken
 * Version:		12
 * Creation Date:	Sat Dec 19 00:47:44 PST 1992
 * Filename:		pointers.h
 * History:
 *	ssl	11	Sun Nov 21 21:19:30 1993
 *		Alternate representation for globPtrType when compiling
 *      with native compiler.
 *	ssl	10	Tue Nov 16 12:03:14 1993
 *		Replaced global tags with macros.
 *	AK	8	Tue Mar 16 14:32:44 1993
 *		Modified num_procs* to handle when procs is explicitly spec.
 *	AK	6	Tue Mar 16 14:32:00 1993
 *		Fixed the bug in ___post_assign_spread_pointer
 * 	TvE	3	Tue Jan 19 00:10:10 PST 1993
 *		PROCS_POWER_OF_TWO changed to ___PROCS_POWER_OF_TWO
 * 	TvE	2	Sat Dec 19 00:48:06 PST 1992
 *		Stuff taken from tic_int.h
 */

#ifndef __POINTERS_H
#define __POINTERS_H

/* Mhh, do we really want to do this? -TvE */
#ifndef NULL
#  define NULL (0)
#endif

/* 
 * Global pointer representation now lives in primitives.h
 */

#include <primitives.h>


/* .............................
 *
 * Constructing global addresses
 *
 * .............................
 */

/* Creates a global pointer which is proc:local */
EXTERN INLINE_MODIFIER jGPointer toglobal(Process proc, void *local)
INLINE({
	jGPointer tempPtr;

	SET_ADDR(tempPtr, local);
	SET_PROC(tempPtr, proc);
	SET_BOX(tempPtr, COMM_GetBoxNumberForProcNumber(proc));
	return tempPtr;
})

/* Creates a global pointer which is proc:local */
EXTERN INLINE_MODIFIER jGPointer toglobalwithbox(Process proc, Process box, 
					void *local)
INLINE({
	jGPointer tempPtr;

	SET_ADDR(tempPtr, local);
	SET_PROC(tempPtr, proc);
	SET_BOX(tempPtr, box);
	return tempPtr;
})

/* Mutates a global pointer which is proc:local */
#define mutateglobal(globPtr, proc, local) \
  do {\
    SET_ADDR(globPtr, local);\
    SET_PROC(globPtr, proc);\
    SET_BOX(globPtr, COMM_GetBoxNumberForProcNumber(proc)); \
} while(0)

/* Mutates a global pointer which is proc:local */
#define mutateglobalwithbox(globPtr, proc, box, local) \
  do {   \
	SET_ADDR(globPtr, local); \
	SET_PROC(globPtr, proc);  \
	SET_BOX(globPtr, box);    \
  } while (0)                       

#if 0
/* don't take a chance on the C inliner */

/* Accessor function to get the processor part */
EXTERN INLINE_MODIFIER Process toproc(jGPointer globPtr)
INLINE({
	return GET_PROC(globPtr);
})

/* Accessor function to get the box part */
EXTERN INLINE_MODIFIER Process tobox(jGPointer globPtr)
INLINE({
	return GET_BOX(globPtr);
})
#else
#define toproc(globPtr) 	GET_PROC(globPtr)
#define tobox(globPtr)		GET_BOX(globPtr)
#endif

#define toboxproc(globPtr) (*(jUIntPointer *)(&globPtr))
#define boxproctoproc(boxproc) (GET_PROC(*(jGPointer *)(&boxproc)))
#define boxproctobox(boxproc)  (GET_BOX(*(jGPointer *)(&boxproc)))

/* Accessor function to get the address part of a global pointer */
     /* Allow tolocal to be an lvalue */
#define tolocal(globPtr)  GET_ADDR(globPtr)

#define islocal(globPtr)   (GET_PROC(globPtr) == MYPROC)


/* ..........................
 *
 * Global pointer comparisons
 *
 * ..........................
 */

EXTERN INLINE_MODIFIER int ___null_ptr_check(jGPointer gPtr)
INLINE ({
	return GET_ADDR(gPtr) == 0;
})

EXTERN INLINE_MODIFIER int ___not_null_ptr_check(jGPointer gPtr)
INLINE ({
	return GET_ADDR(gPtr) != 0;
})

EXTERN INLINE_MODIFIER int ___eq_ptr_check(jGPointer gPtr1, jGPointer gPtr2)
INLINE ({
	return ((GET_ADDR(gPtr1) == GET_ADDR(gPtr2)) &&
	        (GET_PROC(gPtr1) == GET_PROC(gPtr2)))
             || (___null_ptr_check(gPtr1) && ___null_ptr_check(gPtr2));
})

EXTERN INLINE_MODIFIER int ___neq_ptr_check(jGPointer gPtr1, jGPointer gPtr2)
INLINE ({
	return GET_ADDR(gPtr1) != GET_ADDR(gPtr2)
	    || GET_PROC(gPtr1) != GET_PROC(gPtr2);

})


#endif /* __POINTERS_H */
