#ifndef _INCLUDE_TLIB_NATIVE_UTILS_H_
#define _INCLUDE_TLIB_NATIVE_UTILS_H_

#include "linkage.h"
#include "mem-common.h"
#include "primitives.h"

#define _TOSTR_HELP(x) #x
#define _TOSTR(x) _TOSTR_HELP(x)

#ifdef __GNUC__
  /* __func__ should also work for ISO C99 compilers */
  #define __current_loc  __FUNCTION__ "() at " __FILE__ ":" _TOSTR(__LINE__)
#else
  #define __current_loc __FILE__ ":" _TOSTR(__LINE__)
#endif

/* these macros return the value of the expression given, but pass on 
   a hint that you expect the value to be true or false. 
   Use them to wrap the conditional expression in an if stmt when
   you have strong reason to believe the branch will frequently go
   in one direction and the branch is a bottleneck
 */
#ifdef HAVE_BUILTIN_EXPECT
 #define PREDICT_TRUE(exp)  __builtin_expect( (exp), 1 )
 #define PREDICT_FALSE(exp) __builtin_expect( (exp), 0 )
#else
 #define PREDICT_TRUE(exp)  (exp)
 #define PREDICT_FALSE(exp) (exp)
#endif

#define CHECK_NULL_GLOBAL( object )				\
  do {								\
    if (PREDICT_FALSE(isNull(object)))				\
      tossNullPointerException_str(__current_loc);		\
  } while(0)


#define CHECK_NULL_LOCAL( object )				\
  do {								\
    if (PREDICT_FALSE(!(object)))				\
      tossNullPointerException_str(__current_loc);		\
  } while(0)

void tossNullPointerException_str(char * const string) __attribute__((noreturn)
);
 
void tossMonitorStateException_str(char * const string) __attribute__((noreturn));

void tossIllegalArgumentException_str(char * const string) __attribute__((noreturn));


#endif
